from rest_flex_fields import FlexFieldsModelSerializer
from rest_framework import serializers
from rest_framework.fields import CharField, ChoiceField, DateTimeField, DecimalField, IntegerField
from rest_framework.serializers import ModelSerializer, Serializer

from NEMO.models import (
	Account,
	Area,
	AreaAccessRecord,
	Project,
	Reservation,
	Resource,
	ScheduledOutage,
	StaffCharge,
	Task,
	TaskHistory,
	Tool,
	TrainingSession,
	UsageEvent,
	User,
)


class UserSerializer(ModelSerializer):
	class Meta:
		model = User
		fields = (
			"id",
			"type",
			"domain",
			"first_name",
			"last_name",
			"username",
			"email",
			"date_joined",
			"last_login",
			"access_expiration",
			"badge_number",
			"is_active",
			"is_staff",
			"is_superuser",
			"is_technician",
			"is_service_personnel",
			"training_required",
		)


class ProjectSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = Project
		fields = "__all__"
		expandable_fields = {
			"account": "NEMO.serializers.AccountSerializer",
			"only_allow_tools": ("NEMO.serializers.ToolSerializer", {"many": True}),
		}


class AccountSerializer(ModelSerializer):
	class Meta:
		model = Account
		fields = "__all__"


class ToolSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = Tool
		fields = "__all__"
		expandable_fields = {
			"parent_tool": "NEMO.serializers.ToolSerializer",
			"_primary_owner": "NEMO.serializers.UserSerializer",
			"_backup_owners": ("NEMO.serializers.UserSerializer", {"many": True}),
			"_superusers": ("NEMO.serializers.UserSerializer", {"many": True}),
			"_requires_area_access": "NEMO.serializers.AreaSerializer",
			"project": "NEMO.serializers.ProjectSerializer",
			"descendant": "NEMO.serializers.ReservationSerializer",
		}


class AreaSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = Area
		fields = "__all__"
		expandable_fields = {"parent_area": "NEMO.serializers.AreaSerializer"}


class ReservationSerializer(FlexFieldsModelSerializer):
	question_data = serializers.JSONField(source="question_data_json")

	class Meta:
		model = Reservation
		fields = "__all__"
		expandable_fields = {
			"user": "NEMO.serializers.UserSerializer",
			"creator": "NEMO.serializers.UserSerializer",
			"tool": "NEMO.serializers.ToolSerializer",
			"area": "NEMO.serializers.AreaSerializer",
			"project": "NEMO.serializers.ProjectSerializer",
			"descendant": "NEMO.serializers.ReservationSerializer",
		}


class UsageEventSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = UsageEvent
		fields = "__all__"
		expandable_fields = {
			"user": "NEMO.serializers.UserSerializer",
			"operator": "NEMO.serializers.UserSerializer",
			"tool": "NEMO.serializers.ToolSerializer",
			"project": "NEMO.serializers.ProjectSerializer",
		}


class AreaAccessRecordSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = AreaAccessRecord
		fields = "__all__"
		expandable_fields = {
			"customer": "NEMO.serializers.UserSerializer",
			"area": "NEMO.serializers.AreaSerializer",
			"project": "NEMO.serializers.ProjectSerializer",
			"staff_charge": "NEMO.serializers.StaffChargeSerializer",
		}


class TaskHistorySerializer(FlexFieldsModelSerializer):
	class Meta:
		model = TaskHistory
		fields = "__all__"
		expandable_fields = {"user": "NEMO.serializers.UserSerializer", "task": "NEMO.serializers.TaskSerializer"}


class TaskSerializer(FlexFieldsModelSerializer):
	history = TaskHistorySerializer(many=True, read_only=True)

	class Meta:
		model = Task
		fields = "__all__"
		expandable_fields = {
			"tool": "NEMO.serializers.ToolSerializer",
			"creator": "NEMO.serializers.UserSerializer",
			"last_updated_by": "NEMO.serializers.UserSerializer",
			"resolver": "NEMO.serializers.UserSerializer",
		}


class ScheduledOutageSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = ScheduledOutage
		fields = "__all__"
		expandable_fields = {
			"creator": "NEMO.serializers.UserSerializer",
			"tool": "NEMO.serializers.ToolSerializer",
			"area": "NEMO.serializers.AreaSerializer",
			"resource": "NEMO.serializers.ResourceSerializer",
		}


class TrainingSessionSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = TrainingSession
		fields = "__all__"
		expandable_fields = {
			"trainer": "NEMO.serializers.UserSerializer",
			"trainee": "NEMO.serializers.UserSerializer",
			"tool": "NEMO.serializers.ToolSerializer",
			"project": "NEMO.serializers.ProjectSerializer",
		}


class StaffChargeSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = StaffCharge
		fields = "__all__"
		expandable_fields = {
			"customer": "NEMO.serializers.UserSerializer",
			"staff_member": "NEMO.serializers.UserSerializer",
			"project": "NEMO.serializers.ProjectSerializer",
		}


class ResourceSerializer(FlexFieldsModelSerializer):
	class Meta:
		model = Resource
		fields = "__all__"
		expandable_fields = {
			"fully_dependent_tools": ("NEMO.serializers.ToolSerializer", {"many": True}),
			"partially_dependent_tools": ("NEMO.serializers.ToolSerializer", {"many": True}),
			"dependent_areas": ("NEMO.serializers.AreaSerializer", {"many": True}),
		}


class BillableItemSerializer(Serializer):
	type = ChoiceField(
		["missed_reservation", "tool_usage", "area_access", "consumable", "staff_charge", "training_session"]
	)
	name = CharField(max_length=200, read_only=True)
	details = CharField(max_length=500, read_only=True)
	account = CharField(max_length=200, read_only=True)
	account_id = IntegerField(read_only=True)
	project = CharField(max_length=200, read_only=True)
	project_id = IntegerField(read_only=True)
	application = CharField(max_length=200, read_only=True)
	user = CharField(max_length=255, read_only=True)
	username = CharField(max_length=200, read_only=True)
	user_id = IntegerField(read_only=True)
	start = DateTimeField(read_only=True)
	end = DateTimeField(read_only=True)
	quantity = DecimalField(read_only=True, decimal_places=2, max_digits=8)

	def update(self, instance, validated_data):
		pass

	def create(self, validated_data):
		pass

	class Meta:
		fields = "__all__"



from django.core.management import BaseCommand

from NEMO.views.calendar import send_email_usage_reminders


class Command(BaseCommand):
	help = (
		"Run every hour to trigger email usage reminders. "
		"Reservation reminder and reservation warning emails have to be set in customizations for this to work."
	)

	def add_arguments(self, parser):
		parser.add_argument('exclude_project_ids', nargs='*', type=int, help='list of project ids to exclude')

	def handle(self, *args, **options):
		send_email_usage_reminders(options.get('exclude_project_ids'))


# Generated by Django 2.2.10 on 2020-05-21 19:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0016_area_reservation_warning'),
    ]

    operations = [
        migrations.AlterField(
            model_name='configuration',
            name='exclude_from_configuration_agenda',
            field=models.BooleanField(default=False, help_text='Reservations containing this configuration will be excluded from the Configuration Agenda page.'),
        ),
        migrations.AlterField(
            model_name='reservation',
            name='self_configuration',
            field=models.BooleanField(default=False, help_text='When checked, indicates that the user will perform their own tool configuration (instead of requesting that the staff configure it for them).'),
        ),
        migrations.AlterField(
            model_name='reservation',
            name='short_notice',
            field=models.BooleanField(default=None, help_text='Indicates that the reservation was made after the configuration deadline for a tool. Staff may not have enough time to properly configure the tool before the user is scheduled to use it.'),
        ),
        migrations.AlterField(
            model_name='task',
            name='safety_hazard',
            field=models.BooleanField(default=None, help_text='Indicates that this task represents a safety hazard.'),
        ),
        migrations.AlterField(
            model_name='user',
            name='badge_number',
            field=models.PositiveIntegerField(blank=True, help_text='The badge number associated with this user. This number must correctly correspond to a user in order for the tablet-login system (in the lobby) to work properly.', null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='landingpagechoice',
            name='secure_referral',
            field=models.BooleanField(default=True, help_text="Improves security by blocking HTTP referer [sic] information from the targeted page. Enabling this prevents the target page from manipulating the calling page's DOM with JavaScript. This should always be used for external links. It is safe to uncheck this when linking within the site. Leave this box checked if you don't know what this means"),
        ),
        migrations.AlterField(
            model_name='landingpagechoice',
            name='url',
            field=models.CharField(
                help_text='The URL that the choice leads to when clicked. Relative paths such as /calendar/ are used when linking within the site. Use fully qualified URL paths such as https://www.google.com/ to link to external sites.',
                max_length=200, verbose_name='URL'),
        ),
        migrations.AlterField(
            model_name='user',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Designates whether this user can log in. Unselect this instead of deleting accounts.', verbose_name='active'),
        ),
    ]



from django.core.management import BaseCommand

from NEMO.views.calendar import send_email_out_of_time_reservation_notification


class Command(BaseCommand):
	help = (
		"Run every minute to trigger email out of time in area notification. "
		"Out of time reservation email has to be set in customizations for this to work."
	)

	def handle(self, *args, **options):
		send_email_out_of_time_reservation_notification()


# Generated by Django 2.2.13 on 2020-09-17 14:44

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0023_badgereader'),
    ]

    operations = [
        migrations.AddField(
            model_name='contactinformation',
            name='user',
            field=models.OneToOneField(blank=True, help_text='Select a user to associate with this contact. When set, this contact information will be shown instead of the user information on pages like tool details.', null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.apps import AppConfig


def init_admin_site():
	from NEMO.views.customization import get_customization
	from django.contrib import admin
	# customize the site
	site_title = get_customization("site_title", raise_exception=False)
	admin.site.site_header = site_title
	admin.site.site_title = site_title
	admin.site.index_title = "Detailed administration"


def init_rates():
	from NEMO.rates import rate_class
	rate_class.load_rates()


class NEMOConfig(AppConfig):
	name = "NEMO"

	def ready(self):
		from django.apps import apps
		if apps.is_installed("django.contrib.admin"):
			init_admin_site()
		init_rates()


default_app_config = "NEMO.NEMOConfig"



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-01-03 01:22
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0010_version_1_20_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='tool',
            name='parent_tool',
            field=models.ForeignKey(blank=True, help_text='Select a parent tool to allow alternate usage', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tool_children_set', to='NEMO.Tool'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='category',
            field=models.CharField(null=True, blank=True, db_column='category', help_text='Create sub-categories using slashes. For example "Category 1/Sub-category 1".', max_length=1000),
        ),
        migrations.AlterField(
            model_name='tool',
            name='allow_delayed_logoff',
            field=models.BooleanField(db_column='allow_delayed_logoff', default=False,
                                      help_text='Upon logging off users may enter a delay before another user may use the tool. Some tools require "spin-down" or cleaning time after use.'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='backup_owners',
            field=models.ManyToManyField(blank=True, db_table='NEMO_tool_backup_owners',
                                         help_text='Alternate staff members who are responsible for administration of this tool when the primary owner is unavailable.',
                                         related_name='backup_for_tools', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='tool',
            name='grant_badge_reader_access_upon_qualification',
            field=models.CharField(blank=True, db_column='grant_badge_reader_access_upon_qualification',
                                   help_text='Badge reader access is granted to the user upon qualification for this tool.',
                                   max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='grant_physical_access_level_upon_qualification',
            field=models.ForeignKey(blank=True, db_column='grant_physical_access_level_upon_qualification_id',
                                    help_text='The designated physical access level is granted to the user upon qualification for this tool.',
                                    null=True, on_delete=django.db.models.deletion.PROTECT,
                                    to='NEMO.PhysicalAccessLevel'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='interlock',
            field=models.OneToOneField(blank=True, db_column='interlock_id', null=True,
                                       on_delete=django.db.models.deletion.SET_NULL, to='NEMO.Interlock'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='location',
            field=models.CharField(blank=True, db_column='location', max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='maximum_future_reservation_time',
            field=models.PositiveIntegerField(blank=True, db_column='maximum_future_reservation_time',
                                              help_text='The maximum amount of time (in minutes) that a user may reserve from the current time onwards.',
                                              null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='maximum_reservations_per_day',
            field=models.PositiveIntegerField(blank=True, db_column='maximum_reservations_per_day',
                                              help_text='The maximum number of reservations a user may make per day for this tool.',
                                              null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='maximum_usage_block_time',
            field=models.PositiveIntegerField(blank=True, db_column='maximum_usage_block_time',
                                              help_text='The maximum amount of time (in minutes) that a user may reserve this tool for a single reservation. Leave this field blank to indicate that no maximum usage block time exists for this tool.',
                                              null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='minimum_time_between_reservations',
            field=models.PositiveIntegerField(blank=True, db_column='minimum_time_between_reservations',
                                              help_text='The minimum amount of time (in minutes) that the same user must have between any two reservations for this tool.',
                                              null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='minimum_usage_block_time',
            field=models.PositiveIntegerField(blank=True, db_column='minimum_usage_block_time',
                                              help_text='The minimum amount of time (in minutes) that a user must reserve this tool for a single reservation. Leave this field blank to indicate that no minimum usage block time exists for this tool.',
                                              null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='missed_reservation_threshold',
            field=models.PositiveIntegerField(blank=True, db_column='missed_reservation_threshold',
                                              help_text='The amount of time (in minutes) that a tool reservation may go unused before it is automatically marked as "missed" and hidden from the calendar. Usage can be from any user, regardless of who the reservation was originally created for. The cancellation process is triggered by a timed job on the web server.',
                                              null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='notification_email_address',
            field=models.EmailField(blank=True, db_column='notification_email_address',
                                    help_text='Messages that relate to this tool (such as comments, problems, and shutdowns) will be forwarded to this email address. This can be a normal email address or a mailing list address.',
                                    max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='operational',
            field=models.BooleanField(db_column='operational', default=False,
                                      help_text='Marking the tool non-operational will prevent users from using the tool.'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='phone_number',
            field=models.CharField(blank=True, db_column='phone_number', max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='policy_off_between_times',
            field=models.BooleanField(db_column='policy_off_between_times', default=False,
                                      help_text='Check this box to disable policy rules every day between the given times'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='policy_off_end_time',
            field=models.TimeField(blank=True, db_column='policy_off_end_time',
                                   help_text='The end time when policy rules should NOT be enforced', null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='policy_off_start_time',
            field=models.TimeField(blank=True, db_column='policy_off_start_time',
                                   help_text='The start time when policy rules should NOT be enforced', null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='policy_off_weekend',
            field=models.BooleanField(db_column='policy_off_weekend', default=False,
                                      help_text='Whether or not policy rules should be enforced on weekends'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='post_usage_questions',
            field=models.TextField(blank=True, db_column='post_usage_questions', null=True),
        ),
        migrations.AlterField(
            model_name='tool',
            name='primary_owner',
            field=models.ForeignKey(blank=True, db_column='primary_owner_id',
                                    help_text='The staff member who is responsible for administration of this tool.',
                                    null=True, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='primary_tool_owner', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='tool',
            name='requires_area_access',
            field=models.ForeignKey(blank=True, db_column='requires_area_access_id',
                                    help_text='Indicates that this tool is physically located in a billable area and requires an active area access record in order to be operated.',
                                    null=True, on_delete=django.db.models.deletion.PROTECT, to='NEMO.Area'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='reservation_horizon',
            field=models.PositiveIntegerField(blank=True, db_column='reservation_horizon', default=14,
                                              help_text='Users may create reservations this many days in advance. Leave this field blank to indicate that no reservation horizon exists for this tool.',
                                              null=True),
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='category',
            new_name='_category',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='backup_owners',
            new_name='_backup_owners',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='allow_delayed_logoff',
            new_name='_allow_delayed_logoff',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='grant_badge_reader_access_upon_qualification',
            new_name='_grant_badge_reader_access_upon_qualification',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='grant_physical_access_level_upon_qualification',
            new_name='_grant_physical_access_level_upon_qualification',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='interlock',
            new_name='_interlock',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='location',
            new_name='_location',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='maximum_future_reservation_time',
            new_name='_maximum_future_reservation_time',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='maximum_reservations_per_day',
            new_name='_maximum_reservations_per_day',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='maximum_usage_block_time',
            new_name='_maximum_usage_block_time',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='minimum_time_between_reservations',
            new_name='_minimum_time_between_reservations',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='minimum_usage_block_time',
            new_name='_minimum_usage_block_time',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='missed_reservation_threshold',
            new_name='_missed_reservation_threshold',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='notification_email_address',
            new_name='_notification_email_address',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='operational',
            new_name='_operational',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='phone_number',
            new_name='_phone_number',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='policy_off_between_times',
            new_name='_policy_off_between_times',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='policy_off_end_time',
            new_name='_policy_off_end_time',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='policy_off_start_time',
            new_name='_policy_off_start_time',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='policy_off_weekend',
            new_name='_policy_off_weekend',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='post_usage_questions',
            new_name='_post_usage_questions',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='primary_owner',
            new_name='_primary_owner',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='requires_area_access',
            new_name='_requires_area_access',
        ),
        migrations.RenameField(
            model_name='tool',
            old_name='reservation_horizon',
            new_name='_reservation_horizon',
        ),
    ]



import socket
import struct
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Dict
from xml.etree import ElementTree

import requests
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from NEMO.admin import InterlockCardAdminForm, InterlockAdminForm
from NEMO.exceptions import InterlockError
from NEMO.models import Interlock as Interlock_model, InterlockCardCategory
from NEMO.utilities import format_datetime

interlocks_logger = getLogger(__name__)


class Interlock(ABC):
	"""
	This interface allows for customization of Interlock features.
	The only method that has to be implemented is the abstract method "send_command" which send a LOCKED parameter when
	the interlock needs to be locked, and an UNLOCKED parameter when it needs to be unlocked

	The method "clean_interlock_card" can be implemented to set validation rules for the interlock card with the same category

	The interlock type should be set at the end of this file in the dictionary. The key is the key from InterlockCategory, the value is the Interlock implementation.
	"""

	def clean_interlock_card(self, interlock_card_form: InterlockCardAdminForm):
		pass

	def clean_interlock(self, interlock_form: InterlockAdminForm):
		pass

	def lock(self, interlock: Interlock_model) -> {True, False}:
		return self.__issue_command(interlock, Interlock_model.State.LOCKED)

	def unlock(self, interlock: Interlock_model) -> {True, False}:
		return self.__issue_command(interlock, Interlock_model.State.UNLOCKED)

	def __issue_command(self, interlock: Interlock_model, command_type: Interlock_model.State):
		interlocks_enabled = getattr(settings, 'INTERLOCKS_ENABLED', False)
		if not interlocks_enabled or not interlock.card.enabled:
			interlock.most_recent_reply = "Interlock interface mocked out because settings.INTERLOCKS_ENABLED = False or interlock card is disabled. Interlock last set on " + format_datetime(timezone.now()) + "."
			interlock.state = command_type
			interlock.save()
			return True

		state = Interlock_model.State.UNKNOWN
		error_message = ''
		# try to send the command to the interlock
		try:
			state = self._send_command(interlock, command_type)
		except InterlockError as error:
			interlocks_logger.error(error)
			error_message = error.message
		except Exception as error:
			interlocks_logger.error(error)
			error_message = str(error)

		# save interlock state
		interlock.state = state
		interlock.most_recent_reply = Interlock.__create_reply_message(command_type, state, error_message)
		interlock.save()

		# log some useful information
		if interlock.state == interlock.State.UNKNOWN:
			interlocks_logger.error(f"Interlock {interlock.id} is in an unknown state. {interlock.most_recent_reply}")
		elif interlock.state == interlock.State.LOCKED:
			interlocks_logger.debug(f"Interlock {interlock.id} locked successfully at {format_datetime(timezone.now())}")
		elif interlock.state == interlock.State.UNLOCKED:
			interlocks_logger.debug(f"Interlock {interlock.id} unlocked successfully at {format_datetime(timezone.now())}")

		# If the command type equals the current state then the command worked which will return true:
		return interlock.state == command_type

	@staticmethod
	def __create_reply_message(command_type: Interlock_model.State, actual_state: Interlock_model.State, error_message: str) -> str:
		# Compose the status message of the last command.
		reply_message = f"Reply received at {format_datetime(timezone.now())}. "
		if command_type == Interlock_model.State.UNLOCKED:
			reply_message += "Unlock"
		elif command_type == Interlock_model.State.LOCKED:
			reply_message += "Lock"
		else:
			reply_message += "Unknown"
		reply_message += " command "
		if command_type == actual_state and not error_message:
			reply_message += "succeeded."
		else:
			reply_message += "failed. " + error_message
		return reply_message

	@abstractmethod
	def _send_command(self, interlock: Interlock_model, command_type: Interlock_model.State) -> Interlock_model.State:
		pass


class NoOpInterlock(Interlock):

	def _send_command(self, interlock: Interlock_model, command_type: Interlock_model.State) -> Interlock_model.State:
		pass

class StanfordInterlock(Interlock):

	def clean_interlock_card(self, interlock_card_form: InterlockCardAdminForm):
		even_port = interlock_card_form.cleaned_data['even_port']
		odd_port = interlock_card_form.cleaned_data['odd_port']
		number = interlock_card_form.cleaned_data['number']
		error = {}
		if not even_port and even_port != 0:
			error['even_port'] = _('This field is required.')
		if not odd_port and odd_port != 0:
			error['odd_port'] = _('This field is required.')
		if not number and number != 0:
			error['number'] = _('This field is required.')
		if error:
			raise ValidationError(error)

	def clean_interlock(self, interlock_form: InterlockAdminForm):
		channel = interlock_form.cleaned_data['channel']
		error = {}
		if not channel:
			error['channel'] = _('This field is required.')
		if error:
			raise ValidationError(error)

	def _send_command(self, interlock: Interlock_model, command_type: Interlock_model.State) -> Interlock_model.State:
		# The string in this next function call identifies the format of the interlock message.
		# '!' means use network byte order (big endian) for the contents of the message.
		# '20s' means that the message begins with a 20 character string.
		# Each 'i' is an integer field (4 bytes).
		# Each 'b' is a byte field (1 byte).
		# '18s' means that the message ends with a 18 character string.
		# More information on Python structs can be found at:
		# http://docs.python.org/library/struct.html
		command_schema = struct.Struct('!20siiiiiiiiibbbbb18s')
		command_message = command_schema.pack(
			b'EQCNTL_BEGIN_COMMAND',
			1,  # Instruction count
			interlock.card.number,
			interlock.card.even_port,
			interlock.card.odd_port,
			interlock.channel,
			0,  # Command return value
			command_type,  # Type
			0,  # Command
			0,  # Delay
			0,  # SD overload
			0,  # RD overload
			0,  # ADC done
			0,  # Busy
			0,  # Instruction return value
			b'EQCNTL_END_COMMAND'
		)

		# Create a TCP socket to send the interlock command.
		sock = socket.socket()
		try:
			sock.settimeout(3.0)  # Set the send/receive timeout to be 3 seconds.
			server_address = (interlock.card.server, interlock.card.port)
			sock.connect(server_address)
			sock.send(command_message)
			# The reply schema is the same as the command schema except there are no start and end strings.
			reply_schema = struct.Struct('!iiiiiiiiibbbbb')
			reply = sock.recv(reply_schema.size)
			reply = reply_schema.unpack(reply)

			# Update the state of the interlock in the database if the command succeeded.
			if reply[5]:
				interlock_state = command_type
				return interlock_state
			else:
				# raise an exception if it failed
				reply_message = "Stanford Interlock exception:\nResponse information: " +\
								"Instruction count = " + str(reply[0]) + ", " +\
								"card number = " + str(reply[1]) + ", " +\
								"even port = " + str(reply[2]) + ", " +\
								"odd port = " + str(reply[3]) + ", " +\
								"channel = " + str(reply[4]) + ", " +\
								"command return value = " + str(reply[5]) + ", " +\
								"instruction type = " + str(reply[6]) + ", " +\
								"instruction = " + str(reply[7]) + ", " +\
								"delay = " + str(reply[8]) + ", " +\
								"SD overload = " + str(reply[9]) + ", " +\
								"RD overload = " + str(reply[10]) + ", " +\
								"ADC done = " + str(reply[11]) + ", " +\
								"busy = " + str(reply[12]) + ", " +\
								"instruction return value = " + str(reply[13]) + "."
				raise InterlockError(interlock=interlock, msg=reply_message)

		# Log any errors that occurred during the operation into the database.
		except OSError as error:
			reply_message = "Socket error"
			if error.errno:
				reply_message += " " + str(error.errno)
			reply_message += ": " + str(error)
			raise InterlockError(interlock=interlock, msg=reply_message)
		except struct.error as error:
			reply_message = "Response format error: " + str(error)
			raise InterlockError(interlock=interlock, msg=reply_message)
		except InterlockError:
			raise
		except Exception as error:
			reply_message = "General exception: " + str(error)
			raise InterlockError(interlock=interlock, msg=reply_message)
		finally:
			sock.close()


class ProXrInterlock(Interlock):
	"""
	Support for ProXR relay controllers.
	See https://ncd.io/proxr-quick-start-guide/ for more about ProXR.
	"""
	# proxr relay status
	PXR_RELAY_OFF = 0
	PXR_RELAY_ON = 1

	def clean_interlock(self, interlock_form: InterlockAdminForm):
		"""Validates NEMO interlock configuration."""
		channel = interlock_form.cleaned_data['channel']
		error = {}
		if channel not in range(1, 9):
			error['channel'] = _('Relay must be 1-8.')
		if error:
			raise ValidationError(error)

	def _send_bytes(self, relay_socket, proxrcmd):
		"""
		Returns the response from the relay controller.
		Argument relay_socket is a connected socket object.
		Argument proxrcmd (the ProXR command) is an iterable of 8-bit integers.
		"""
		# make sure that all bytes get sent
		bytes_sent = 0
		while bytes_sent < len(proxrcmd):
			bytes_sent += relay_socket.send(bytes(proxrcmd[bytes_sent:]))
		# relay responses can include erroneous data
		# only the last byte of the response is important
		return relay_socket.recv(64)[-1]

	def _get_state(self, relay_socket, interlock_channel):
		"""
		Returns current NEMO state of the relay.
		Argument relay_socket is a connected socket object.
		Argument interlock_channel is the NEMO interlock.channel.
		"""
		state = self._send_bytes(relay_socket, (254, 115 + interlock_channel, 1))
		if state == self.PXR_RELAY_OFF:
			return Interlock_model.State.LOCKED
		elif state == self.PXR_RELAY_ON:
			return Interlock_model.State.UNLOCKED
		else:
			return Interlock_model.State.UNKNOWN

	def _send_command(self, interlock: Interlock_model, command_type: Interlock_model.State) -> Interlock_model.State:
		"""Returns and sets NEMO locked/unlocked state."""
		state = Interlock_model.State.UNKNOWN
		try:
			with socket.create_connection((interlock.card.server, interlock.card.port), 10) as relay_socket:
				if command_type == Interlock_model.State.LOCKED:
					# turn the interlock channel off
					self._send_bytes(relay_socket, (254, 99 + interlock.channel, 1))
					state = self._get_state(relay_socket, interlock.channel)
				elif command_type == Interlock_model.State.UNLOCKED:
					# turn the interlock channel on
					self._send_bytes(relay_socket, (254, 107 + interlock.channel, 1))
					state = self._get_state(relay_socket, interlock.channel)
		except Exception as error:
			raise InterlockError(interlock=interlock, msg="Communication error: " + str(error))
		return state


class WebRelayHttpInterlock(Interlock):
	WEB_RELAY_OFF = 0
	WEB_RELAY_ON = 1

	def clean_interlock_card(self, interlock_card_form: InterlockCardAdminForm):
		username = interlock_card_form.cleaned_data['username']
		password = interlock_card_form.cleaned_data['password']
		error = {}
		if username and not password:
			error['password'] = _('password is required when using a username.')
		if error:
			raise ValidationError(error)

	def _send_command(self, interlock: Interlock_model, command_type: Interlock_model.State) -> Interlock_model.State:
		state = Interlock_model.State.UNKNOWN
		try:
			if command_type == Interlock_model.State.LOCKED:
				state = WebRelayHttpInterlock.setRelayState(interlock, WebRelayHttpInterlock.WEB_RELAY_OFF)
			elif command_type == Interlock_model.State.UNLOCKED:
				state = WebRelayHttpInterlock.setRelayState(interlock, WebRelayHttpInterlock.WEB_RELAY_ON)
		except Exception as error:
			raise InterlockError(interlock=interlock, msg="General exception: " + str(error))
		return state

	@staticmethod
	def setRelayState(interlock: Interlock_model, state: {0, 1}) -> Interlock_model.State:
		url = f"{interlock.card.server}:{interlock.card.port}/stateFull.xml?relay{interlock.channel}State={state}"
		if not url.startswith('http') and not url.startswith('https'):
			url = 'http://' + url
		auth = None
		if interlock.card.username and interlock.card.password:
			auth = (interlock.card.username, interlock.card.password)
		response = requests.get(url, auth=auth)
		response.raise_for_status()
		responseXML = ElementTree.fromstring(response.content)
		state = int(responseXML.find(f"relay{interlock.channel}state").text)
		if state == WebRelayHttpInterlock.WEB_RELAY_OFF:
			return Interlock_model.State.LOCKED
		elif state == WebRelayHttpInterlock.WEB_RELAY_ON:
			return Interlock_model.State.UNLOCKED
		else:
			raise Exception(f"Unexpected state received from interlock: {state}")


def get(category: InterlockCardCategory, raise_exception=True):
	"""	Returns the corresponding interlock implementation, and raises an exception if not found. """
	interlock_impl = interlocks.get(category.key, False)
	if not interlock_impl:
		if raise_exception:
			raise Exception(f"There is no interlock implementation for category: {category.name}. Please add one in interlocks.py")
		else:
			return NoOpInterlock()
	else:
		return interlock_impl


interlocks: Dict[str, Interlock] = {
	'stanford': StanfordInterlock(),
	'web_relay_http': WebRelayHttpInterlock(),
	'proxr': ProXrInterlock(),
}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-10-08 00:58
from __future__ import unicode_literals

from django.conf import settings

import NEMO.utilities
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0008_version_1_16_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaskImages',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(upload_to=NEMO.utilities.get_task_image_filename, verbose_name='Image')),
                ('uploaded_at', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'verbose_name_plural': 'Task images',
                'ordering': ['-uploaded_at'],
            },
        ),
        migrations.AlterField(
            model_name='alert',
            name='creator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='comment',
            name='hidden_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='hidden_comments', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='door',
            name='area',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='doors', to='NEMO.Area'),
        ),
        migrations.AlterField(
            model_name='reservation',
            name='cancelled_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='resource',
            name='category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='NEMO.ResourceCategory'),
        ),
        migrations.AlterField(
            model_name='safetyissue',
            name='reporter',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='reported_safety_issues', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='safetyissue',
            name='resolver',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='resolved_safety_issues', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='task',
            name='last_updated_by',
            field=models.ForeignKey(blank=True,
                                    help_text='The last user who modified this task. This should always be a staff member.',
                                    null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='task',
            name='problem_category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='problem_category', to='NEMO.TaskCategory'),
        ),
        migrations.AlterField(
            model_name='task',
            name='resolution_category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='resolution_category', to='NEMO.TaskCategory'),
        ),
        migrations.AlterField(
            model_name='task',
            name='resolver',
            field=models.ForeignKey(blank=True, help_text='The staff member who resolved the task.', null=True,
                                    on_delete=django.db.models.deletion.SET_NULL, related_name='task_resolver',
                                    to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='tool',
            name='primary_owner',
            field=models.ForeignKey(help_text='The staff member who is responsible for administration of this tool.',
                                    on_delete=django.db.models.deletion.PROTECT, related_name='primary_tool_owner',
                                    to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='tool',
            name='requires_area_access',
            field=models.ForeignKey(blank=True,
                                    help_text='Indicates that this tool is physically located in a billable area and requires an active area access record in order to be operated.',
                                    null=True, on_delete=django.db.models.deletion.PROTECT, to='NEMO.Area'),
        ),
        migrations.AddField(
            model_name='taskimages',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Task'),
        ),
    ]



from typing import List, Optional

from django.contrib.auth.decorators import login_required
from django.db.models import Q, QuerySet, Count, Prefetch
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.http import require_GET

from NEMO.decorators import disable_session_expiry_refresh
from NEMO.model_tree import get_area_model_tree, TreeItem, ModelTreeHelper
from NEMO.models import AreaAccessRecord, Resource, ScheduledOutage, Task, Tool, UsageEvent, User, Area
from NEMO.views.customization import get_customization


@login_required
@require_GET
@disable_session_expiry_refresh
def status_dashboard(request, tab=None):
	"""
	Present a web page to allow users to view the status and usage of all tools.
	"""
	interest = request.GET.get('interest')
	if interest is None:
		area_items, no_occupants = process_area_access_record_with_parents(request.user)
		dictionary = {
			'tab': tab if tab else "occupancy",
			'tool_summary': create_tool_summary(),
			'area_items': area_items,
			'no_occupants': no_occupants,
		}
		return render(request, 'status_dashboard/status_dashboard.html', dictionary)
	elif interest == "tools":
		dictionary = {
			'tool_summary': create_tool_summary(),
		}
		return render(request, 'status_dashboard/tools.html', dictionary)
	elif interest == "occupancy":
		area_items, no_occupants = process_area_access_record_with_parents(request.user)
		dictionary = {
			'area_items': area_items,
			'no_occupants': no_occupants,
		}
		return render(request, 'status_dashboard/occupancy.html', dictionary)


def process_area_access_record_with_parents(user: User):
	show_not_qualified_areas = get_customization('dashboard_display_not_qualified_areas')
	records = AreaAccessRecord.objects.filter(end=None, staff_charge=None)
	if not user.is_staff and show_not_qualified_areas != 'enabled':
		records = records.filter(area__in=user.accessible_areas())
	records = records.prefetch_related('customer', 'project', 'area')
	no_occupants = not records.exists()
	area_items = None
	area_model_tree = get_area_model_tree()
	if not no_occupants:
		areas_and_parents = area_model_tree.get_ancestor_areas(area_model_tree.get_areas([record.area.id for record in records]), include_self=True)
		# Sort to have area without children before others
		areas_and_parents.sort(key=lambda x: f'{x.tree_category}zz' if x.is_leaf else f'{x.tree_category}/aa')
		area_summary = create_area_summary(area_model_tree=area_model_tree, add_resources=False, add_outages=False, add_occupants=True)
		area_summary_dict = {area['id']: area for area in area_summary}
		for area_item in areas_and_parents:
			area_item.item = area_summary_dict[area_item.id]
		area_items = area_tree_helper(areas_and_parents, records)
	return area_items, no_occupants


def area_tree_helper(filtered_area: List[TreeItem], records: QuerySet, areas: Optional[List[TreeItem]] = None):
	""" Recursively build a list of areas. The resulting list is meant to be iterated over in a view """
	if areas is None:
		# Get the root areas
		areas = [area for area in filtered_area if area.is_root]
	else:
		yield 'in'

	for area in areas:
		yield area
		children = [child for child in area.children if child in filtered_area]
		if len(children):
			area.leaf = False
			for x in area_tree_helper(filtered_area, records, children):
				yield x
		else:
			area.occupants = records.filter(area__id=area.id)
			area.leaf = True
	yield 'out'


def create_tool_summary():
	tools = Tool.objects.filter(visible=True).prefetch_related(Prefetch('_requires_area_access', queryset=Area.objects.all().only('name')))
	tasks = Task.objects.filter(cancelled=False, resolved=False, tool__visible=True).prefetch_related('tool')
	unavailable_resources = Resource.objects.filter(available=False).prefetch_related('fully_dependent_tools', 'partially_dependent_tools')
	# also check for visibility on the parent if there is one (alternate tool are hidden)
	usage_events = UsageEvent.objects.filter(Q(end=None, tool__visible=True)|Q(end=None, tool__parent_tool__visible=True)).prefetch_related('operator', 'user', 'tool')
	scheduled_outages = ScheduledOutage.objects.filter(start__lte=timezone.now(), end__gt=timezone.now(), area__isnull=True)
	tool_summary = merge(tools, tasks, unavailable_resources, usage_events, scheduled_outages)
	tool_summary = list(tool_summary.values())
	tool_summary.sort(key=lambda x: x['name'])
	return tool_summary


def create_area_summary(area_model_tree: ModelTreeHelper=None, add_resources=True, add_occupants=True, add_outages=True):
	if area_model_tree is None:
		area_model_tree = get_area_model_tree()
	area_items = area_model_tree.items.values()
	result = {}
	for area in area_items:
		result[area.id] = {
			'name': area.name,
			'id': area.id,
			'maximum_capacity': area.maximum_capacity,
			'warning_capacity': area.item.warning_capacity(),
			'danger_capacity': area.item.danger_capacity(),
			'count_staff_in_occupancy': area.count_staff_in_occupancy,
			'count_service_personnel_in_occupancy': area.count_service_personnel_in_occupancy,
			'occupancy_count': 0,
			'occupancy': 0,
			'occupancy_staff': 0,
			'occupancy_service_personnel': 0,
			'occupants': '',
			'required_resource_is_unavailable': False,
			'scheduled_outage': False,
		}

	if add_resources:
		unavailable_resources = Resource.objects.filter(available=False).prefetch_related(Prefetch('dependent_areas', queryset=Area.objects.only('id')))
		for resource in unavailable_resources:
			for area in resource.dependent_areas.all():
				if area.id in result:
					result[area.id]['required_resource_is_unavailable'] = True
	if add_outages:
		scheduled_outages = ScheduledOutage.objects.filter(start__lte=timezone.now(), end__gt=timezone.now(), tool__isnull=True).only('area_id', 'resource_id')
		for outage in scheduled_outages:
			if outage.area_id:
				result[outage.area_id]['scheduled_outage'] = True
			elif outage.resource_id:
				for t in outage.resource.dependent_areas.values_list('id', flat=True):
					result[t]['scheduled_outage'] = True

	if add_occupants:
		occupants: List[AreaAccessRecord] = AreaAccessRecord.objects.filter(end=None, staff_charge=None).prefetch_related(Prefetch('customer', queryset=User.objects.all().only('first_name', 'last_name', 'username', 'is_staff')))
		for occupant in occupants:
			# Get ids for area and all the parents (so we can add occupants info on parents)
			area_ids = area_model_tree.get_area(occupant.area_id).ancestor_ids(include_self=True)
			if occupant.customer.is_staff:
				customer_display = f'<span class="success-highlight">{str(occupant.customer)}</span>'
			elif occupant.customer.is_service_personnel:
				customer_display = f'<span class="warning-highlight">{str(occupant.customer)}</span>'
			elif occupant.customer.is_logged_in_area_without_reservation():
				customer_display = f'<span class="danger-highlight">{str(occupant.customer)}</span>'
			else:
				customer_display = str(occupant.customer)
			for area_id in area_ids:
				if area_id in result:
					result[area_id]['occupancy'] += 1
					if occupant.customer.is_staff:
						result[area_id]['occupancy_staff'] += 1
					if occupant.customer.is_service_personnel:
						result[area_id]['occupancy_service_personnel'] += 1
					if (not occupant.customer.is_staff or result[area_id]['count_staff_in_occupancy']) and (not occupant.customer.is_service_personnel or result[area_id]['count_service_personnel_in_occupancy']):
						result[area_id]['occupancy_count'] += 1
					result[area_id]['occupants'] += customer_display if not result[area_id]['occupants'] else f'<br>{customer_display}'
	area_summary = list(result.values())
	area_summary.sort(key=lambda x: x['name'])
	return area_summary


def merge(tools, tasks, unavailable_resources, usage_events, scheduled_outages):
	result = {}
	tools_with_delayed_logoff_in_effect = [x.tool.tool_or_parent_id() for x in UsageEvent.objects.filter(end__gt=timezone.now())]
	parent_ids = Tool.objects.filter(parent_tool__isnull=False).values_list('parent_tool_id', flat=True)
	for tool in tools:
		result[tool.tool_or_parent_id()] = {
			'name': tool.name_or_child_in_use_name(parent_ids=parent_ids),
			'id': tool.id,
			'user': '',
			'operator': '',
			'in_use': False,
			'in_use_since': '',
			'delayed_logoff_in_progress': tool.tool_or_parent_id() in tools_with_delayed_logoff_in_effect,
			'problematic': False,
			'operational': tool.operational,
			'required_resource_is_unavailable': False,
			'nonrequired_resource_is_unavailable': False,
			'scheduled_outage': False,
			'scheduled_partial_outage': False,
			'area_name': tool.requires_area_access.name if tool.requires_area_access else None,
			'area_requires_reservation': tool.requires_area_access.requires_reservation if tool.requires_area_access else False,
		}
	for task in tasks:
		result[task.tool.id]['problematic'] = True
	for event in usage_events:
		result[event.tool.tool_or_parent_id()]['operator'] = str(event.operator)
		result[event.tool.tool_or_parent_id()]['user'] = str(event.operator)
		if event.user != event.operator:
			result[event.tool.tool_or_parent_id()]['user'] += " on behalf of " + str(event.user)
		result[event.tool.tool_or_parent_id()]['in_use'] = True
		result[event.tool.tool_or_parent_id()]['in_use_since'] = event.start
	for resource in unavailable_resources:
		for tool in resource.fully_dependent_tools.filter(visible=True):
			result[tool.id]['required_resource_is_unavailable'] = True
		for tool in resource.partially_dependent_tools.filter(visible=True):
			result[tool.id]['nonrequired_resource_is_unavailable'] = True
	for outage in scheduled_outages:
		if outage.tool_id and outage.tool.visible:
			result[outage.tool.id]['scheduled_outage'] = True
		elif outage.resource_id:
			for t in outage.resource.fully_dependent_tools.filter(visible=True):
				result[t.id]['scheduled_outage'] = True
			for t in outage.resource.partially_dependent_tools.filter(visible=True):
				result[t.id]['scheduled_partial_outage'] = True
	return result



import csv
import os
from calendar import monthrange
from datetime import timedelta, datetime
from email import encoders
from email.mime.base import MIMEBase
from io import BytesIO
from typing import Tuple, List, Dict, Set

from PIL import Image
from dateutil import parser
from dateutil.parser import parse
from dateutil.rrule import MONTHLY, rrule
from django.core.files.base import ContentFile
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.mail import EmailMessage
from django.db.models import QuerySet
from django.http import HttpResponse
from django.utils import timezone
from django.utils.timezone import localtime


class BasicDisplayTable(object):
	""" Utility table to make adding headers and rows easier, and export to csv """

	def __init__(self):
		# headers is a list of tuples (key, display)
		self.headers: List[Tuple[str, str]] = []
		# rows is a list of dictionaries. Each dictionary is a row, with keys corresponding to header keys
		self.rows: List[Dict] = []

	def add_header(self, header: Tuple[str, str]):
		if not any(k[0] == header[0] for k in self.headers):
			self.headers.append((header[0], header[1].capitalize()))

	def add_row(self, row: Dict):
		self.rows.append(row)

	def flat_headers(self) -> List[str]:
		return [display for key, display in self.headers]

	def flat_rows(self) -> List[List]:
		flat_result = []
		for row in self.rows:
			flat_result.append([row.get(key, "") for key, display_value in self.headers])
		return flat_result

	def to_csv(self) -> HttpResponse:
		response = HttpResponse(content_type="text/csv")
		writer = csv.writer(response)
		writer.writerow([display_value.capitalize() for key, display_value in self.headers])
		for row in self.rows:
			writer.writerow([row.get(key, "") for key, display_value in self.headers])
		return response


def bootstrap_primary_color(color_type):
	if color_type == "success":
		return "#5cb85c"
	elif color_type == "info":
		return "#5bc0de"
	elif color_type == "warning":
		return "#f0ad4e"
	elif color_type == "danger":
		return "#d9534f"
	return None


class EmailCategory(object):
	GENERAL = 0
	SYSTEM = 1
	DIRECT_CONTACT = 2
	BROADCAST_EMAIL = 3
	TIMED_SERVICES = 4
	FEEDBACK = 5
	ABUSE = 6
	SAFETY = 7
	TASKS = 8
	Choices = (
		(GENERAL, "General"),
		(SYSTEM, "System"),
		(DIRECT_CONTACT, "Direct Contact"),
		(BROADCAST_EMAIL, "Broadcast Email"),
		(TIMED_SERVICES, "Timed Services"),
		(FEEDBACK, "Feedback"),
		(ABUSE, "Abuse"),
		(SAFETY, "Safety"),
		(TASKS, "Tasks"),
	)


def parse_start_and_end_date(start, end):
	start = timezone.make_aware(parser.parse(start), timezone.get_current_timezone())
	end = timezone.make_aware(parser.parse(end), timezone.get_current_timezone())
	end += timedelta(days=1, seconds=-1)  # Set the end date to be midnight by adding a day.
	return start, end


def quiet_int(value_to_convert, default_upon_failure=0):
	"""
	Attempt to convert the given value to an integer. If there is any problem
	during the conversion, simply return 'default_upon_failure'.
	"""
	result = default_upon_failure
	try:
		result = int(value_to_convert)
	except:
		pass
	return result


def parse_parameter_string(
		parameter_dictionary, parameter_key, maximum_length=3000, raise_on_error=False, default_return=""
):
	"""
	Attempts to parse a string from an HTTP GET or POST dictionary and applies validation checks.
	"""
	try:
		parameter = parameter_dictionary[parameter_key].strip()
		if raise_on_error and len(parameter) > maximum_length:
			raise Exception(
				"The parameter named {} is {} characters long, exceeding the maximum length of {} characters.".format(
					parameter_key, len(parameter), maximum_length
				)
			)
		return parameter
	except Exception as e:
		if raise_on_error:
			raise e
		return default_return


def month_list(since=datetime(year=2013, month=11, day=1)):
	month_count = (timezone.now().year - since.year) * 12 + (timezone.now().month - since.month) + 1
	result = list(rrule(MONTHLY, dtstart=since, count=month_count))
	result = localize(result)
	result.reverse()
	return result


def get_month_timeframe(date=None):
	if date:
		start = parse(date)
	else:
		start = timezone.now()
	first_of_the_month = localize(datetime(start.year, start.month, 1))
	last_of_the_month = localize(
		datetime(start.year, start.month, monthrange(start.year, start.month)[1], 23, 59, 59, 999999)
	)
	return first_of_the_month, last_of_the_month


def extract_times(parameters, input_timezone=None, start_required=True, end_required=True) -> Tuple[datetime, datetime]:
	"""
	Extract the "start" and "end" parameters from an HTTP request while performing a few logic validation checks.
	The function assumes the UNIX timestamp is in the local timezone. Use input_timezone to specify the timezone.
	"""
	start, end, new_start, new_end = None, None, None, None
	try:
		start = parameters["start"]
	except:
		if start_required:
			raise Exception("The request parameters did not contain a start time.")

	try:
		end = parameters["end"]
	except:
		if end_required:
			raise Exception("The request parameters did not contain an end time.")

	try:
		new_start = float(start)
		new_start = datetime.utcfromtimestamp(new_start)
		new_start = localize(new_start, input_timezone)
	except:
		if start or start_required:
			raise Exception("The request parameters did not have a valid start time.")

	try:
		new_end = float(end)
		new_end = datetime.utcfromtimestamp(new_end)
		new_end = localize(new_end, input_timezone)
	except:
		if end or end_required:
			raise Exception("The request parameters did not have a valid end time.")

	if start and end and start_required and end_required and new_end < new_start:
		raise Exception("The request parameters have an end time that precedes the start time.")

	return new_start, new_end


def extract_date(date):
	return localize(datetime.strptime(date, "%Y-%m-%d"))


def extract_dates(parameters):
	"""
	Extract the "start" and "end" parameters from an HTTP request while performing a few logic validation checks.
	"""
	try:
		start = parameters["start"]
	except:
		raise Exception("The request parameters did not contain a start time.")

	try:
		end = parameters["end"]
	except:
		raise Exception("The request parameters did not contain an end time.")

	try:
		start = extract_date(start)
	except:
		raise Exception("The request parameters did not have a valid start time.")

	try:
		end = extract_date(end)
	except:
		raise Exception("The request parameters did not have a valid end time.")

	if end < start:
		raise Exception("The request parameters have an end time that precedes the start time.")

	return start, end


def format_datetime(universal_time):
	local_time = universal_time.astimezone(timezone.get_current_timezone())
	day = int(local_time.strftime("%d"))
	if 4 <= day <= 20 or 24 <= day <= 30:
		suffix = "th"
	else:
		suffix = ["st", "nd", "rd"][day % 10 - 1]
	return (
			local_time.strftime("%A, %B ")
			+ str(day)
			+ suffix
			+ local_time.strftime(", %Y @ ")
			+ local_time.strftime("%I:%M %p").lstrip("0")
	)


def localize(dt, tz=None):
	tz = tz or timezone.get_current_timezone()
	if isinstance(dt, list):
		return [tz.localize(d) for d in dt]
	else:
		return tz.localize(dt)


def naive_local_current_datetime():
	return localtime(timezone.now()).replace(tzinfo=None)


def beginning_of_the_day(t: datetime, in_local_timezone=True) -> datetime:
	""" Returns the BEGINNING of today's day (12:00:00.000000 AM of the current day) in LOCAL time. """
	midnight = t.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
	return localize(midnight) if in_local_timezone else midnight


def end_of_the_day(t: datetime, in_local_timezone=True) -> datetime:
	""" Returns the END of today's day (11:59:59.999999 PM of the current day) in LOCAL time. """
	midnight = t.replace(hour=23, minute=59, second=59, microsecond=999999, tzinfo=None)
	return localize(midnight) if in_local_timezone else midnight


def send_mail(subject, content, from_email, to=None, bcc=None, cc=None, attachments=None, email_category:EmailCategory = EmailCategory.GENERAL, fail_silently=True) -> int:
	mail = EmailMessage(
		subject=subject, body=content, from_email=from_email, to=to, bcc=bcc, cc=cc, attachments=attachments
	)
	mail.content_subtype = "html"
	msg_sent = 0
	if mail.recipients():
		email_record = create_email_log(mail, email_category)
		try:
			msg_sent = mail.send()
		except:
			email_record.ok = False
			if not fail_silently:
				raise
		finally:
			email_record.save()
	return msg_sent


def create_email_log(email: EmailMessage, email_category: EmailCategory):
	from NEMO.models import EmailLog
	email_record: EmailLog = EmailLog.objects.create(category=email_category, sender=email.from_email, to=', '.join(email.recipients()), subject=email.subject, content=email.body)
	if email.attachments:
		email_attachments = []
		for attachment in email.attachments:
			if isinstance(attachment, MIMEBase):
				email_attachments.append(attachment.get_filename() or '')
		if email_attachments:
			email_record.attachments = ', '.join(email_attachments)
	return email_record


def create_email_attachment(stream, filename=None, maintype="application", subtype="octet-stream", **content_type_params) -> MIMEBase:
	attachment = MIMEBase(maintype, subtype, **content_type_params)
	attachment.set_payload(stream.read())
	encoders.encode_base64(attachment)
	if filename:
		attachment.add_header("Content-Disposition", f'attachment; filename="{filename}"')
	return attachment


def get_task_image_filename(task_images, filename):
	from NEMO.models import Task, TaskImages
	from django.template.defaultfilters import slugify

	task: Task = task_images.task
	tool_name = slugify(task.tool)
	now = datetime.now()
	date = now.strftime("%Y-%m-%d")
	year = now.strftime("%Y")
	number = "{:02d}".format(
		TaskImages.objects.filter(
			task__tool=task.tool, uploaded_at__year=now.year, uploaded_at__month=now.month, uploaded_at__day=now.day
		).count()
		+ 1
	)
	ext = os.path.splitext(filename)[1]
	return f"task_images/{year}/{tool_name}/{date}_{tool_name}_{number}{ext}"


def get_tool_image_filename(tool, filename):
	from django.template.defaultfilters import slugify

	tool_name = slugify(tool)
	ext = os.path.splitext(filename)[1]
	return f"tool_images/{tool_name}{ext}"


def get_tool_document_filename(tool_documents, filename):
	from django.template.defaultfilters import slugify

	tool_name = slugify(tool_documents.tool)
	return f"tool_documents/{tool_name}/{filename}"


def resize_image(image: InMemoryUploadedFile, max: int, quality=85) -> InMemoryUploadedFile:
	""" Returns a resized image based on the given maximum size """
	with Image.open(image) as img:
		width, height = img.size
		# no need to resize if width or height is already less than the max
		if width <= max or height <= max:
			return image
		if width > height:
			width_ratio = max / float(width)
			new_height = int((float(height) * float(width_ratio)))
			img = img.resize((max, new_height), Image.ANTIALIAS)
		else:
			height_ratio = max / float(height)
			new_width = int((float(width) * float(height_ratio)))
			img = img.resize((new_width, max), Image.ANTIALIAS)
		with BytesIO() as buffer:
			img.save(fp=buffer, format="PNG", quality=quality)
			resized_image = ContentFile(buffer.getvalue())
	file_name_without_ext = os.path.splitext(image.name)[0]
	return InMemoryUploadedFile(
		resized_image, "ImageField", "%s.png" % file_name_without_ext, "image/png", resized_image.tell(), None
	)


def distinct_qs_value_list(qs: QuerySet, field_name:str) -> Set:
	return set(list(qs.values_list(field_name, flat=True)))



# Generated by Django 2.2.10 on 2020-04-30 17:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0014_version_2_2_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='area',
            name='maximum_capacity',
            field=models.PositiveIntegerField(default=0, help_text='The maximum number of people allowed in this area at any given time. Set to 0 for unlimited.'),
        ),
    ]



from typing import List, Any

from NEMO.models import User, Area, Resource, Interlock, PhysicalAccessException


class NEMOException(Exception):
	""" Basic NEMO exception """

	default_msg = "A {} error occurred"

	def __init__(self, msg=None):
		if msg is None:
			from NEMO.views.customization import get_customization
			site_title = get_customization('site_title')
			msg = self.default_msg.format(site_title)
		self.msg = msg
		super().__init__(msg)


class InvalidCustomizationException(NEMOException):
	def __init__(self, name: str, value: str = None):
		msg = f"Invalid customization ({name})"
		if value is not None:
			msg += f" for value: [{value}]"
		super().__init__(msg)


class InterlockError(NEMOException):
	""" Interlock related errors """

	def __init__(self, interlock: Interlock, msg=None):
		self.message = msg
		error_message = f"An Error occurred with interlock [{interlock}]"
		if msg is not None:
			error_message += f": {msg}"
		self.interlock = interlock
		super().__init__(error_message)


class UserAccessError(NEMOException):
	""" User access related errors """

	detailed_msg = ""

	def __init__(self, user: Any, msg=None):
		message = f"An Error occurred with user access [{user}]"
		if msg is not None:
			message += f": {msg}"
		elif self.detailed_msg:
			message += f": {self.detailed_msg}"
		self.user = user
		super().__init__(message)


class InactiveUserError(UserAccessError):
	detailed_msg = "This user is not active"


class NoActiveProjectsForUserError(UserAccessError):
	detailed_msg = "This user does not have any active projects"


class PhysicalAccessExpiredUserError(UserAccessError):
	detailed_msg = "This user's physical access has expired"


class NoPhysicalAccessUserError(UserAccessError):
	detailed_msg = "This user has not been granted physical access to any area"


class NoAccessiblePhysicalAccessUserError(UserAccessError):
	def __init__(self, user: User, area: Area, access_exception: PhysicalAccessException = None):
		self.area = area
		self.access_exception = access_exception
		details = f"This user is not assigned to a physical access that allow access to this area [{area}] at this time"
		if self.access_exception:
			details = f"This user was denied access to this area [{area}] due to the following exception: {access_exception.name}"
		super().__init__(user=user, msg=details)


class UnavailableResourcesUserError(UserAccessError):
	def __init__(self, user: User, area: Area, resources: List[Resource]):
		self.area = area
		self.resources = resources
		details = f"This user was denied access to this area [{area}] because a required resource was unavailable [{resources}"
		super().__init__(user=user, msg=details)


class MaximumCapacityReachedError(UserAccessError):
	def __init__(self, user: User, area: Area):
		self.area = area
		details = f"This user was denied access to this area [{area}] because the area's maximum capacity of [{area.maximum_capacity}] has been reached"
		super().__init__(user=user, msg=details)


class ScheduledOutageInProgressError(UserAccessError):
	def __init__(self, user: User, area: Area):
		self.area = area
		details = f"This user was denied access to this area [{area}] because there is a scheduled outage in progress"
		super().__init__(user=user, msg=details)


class ReservationRequiredUserError(UserAccessError):
	def __init__(self, user: User, area: Area):
		self.area = area
		details = f"This user was denied access to this area [{area}] because the user doesn't have a current reservation for that area"
		super().__init__(user=user, msg=details)


# Project policy errors
class ProjectChargeException(NEMOException):
	def __init__(self, project, user, msg=None):
		self.project = project
		self.user = user
		super().__init__(msg)


class NotAllowedToChargeProjectException(ProjectChargeException):
	def __init__(self, project, user, msg=None):
		new_msg = f"Permission to bill project {project.name} was denied."
		super().__init__(project, user, msg or new_msg)


class ItemNotAllowedForProjectException(ProjectChargeException):
	def __init__(self, project, user, item_name, msg=None):
		self.item_name = item_name
		super().__init__(project, user, msg)


class RequiredUnansweredQuestionsException(NEMOException):
	def __init__(self, run_data:str, questions: List):
		self.run_data = run_data
		self.questions = questions
		display_questions = ", ".join([f"\"{question.title}\"" for question in questions])
		msg = f"You have to answer the following required questions: {display_questions}"
		super().__init__(msg)



from datetime import timedelta
from html.parser import HTMLParser
from logging import getLogger

from django.contrib.auth.decorators import login_required
from django.db.models import F
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import localize
from django.utils.http import urlencode
from django.views.decorators.http import require_GET, require_POST, require_http_methods
from mptt.forms import TreeNodeChoiceField

from NEMO.decorators import disable_session_expiry_refresh, staff_member_required, synchronized
from NEMO.exceptions import (
	InactiveUserError,
	MaximumCapacityReachedError,
	NoAccessiblePhysicalAccessUserError,
	NoActiveProjectsForUserError,
	NoPhysicalAccessUserError,
	PhysicalAccessExpiredUserError,
	ProjectChargeException,
	ReservationRequiredUserError,
	ScheduledOutageInProgressError,
	UnavailableResourcesUserError,
)
from NEMO.models import Area, AreaAccessRecord, Project, User
from NEMO.utilities import parse_start_and_end_date, quiet_int
from NEMO.views.calendar import shorten_reservation
from NEMO.views.customization import get_customization
from NEMO.views.policy import check_billing_to_project, check_policy_to_enter_any_area, check_policy_to_enter_this_area

area_access_logger = getLogger(__name__)


# Utility parser to find error messages in rendered self_login view
class ParseSelfLoginErrorMessage(HTMLParser):

	record = False
	data = None

	def handle_starttag(self, startTag, attrs):
		if startTag == "div" and ('class', 'alert alert-danger') in attrs:
			self.record = True

	def handle_data(self, data):
		if self.record:
			self.data = data.strip()

	def handle_endtag(self, endTag):
		if self.record and endTag == "div":
			self.record = False

	def error(self, message):
		pass


@staff_member_required
@require_GET
def area_access(request):
	""" Presents a page that displays audit records for all areas. """
	now = timezone.now().astimezone()
	today = now.strftime('%m/%d/%Y')
	yesterday = (now - timedelta(days=1)).strftime('%m/%d/%Y')
	area_id = ''
	dictionary = {
		'today': reverse('area_access') + '?' + urlencode({'start': today, 'end': today}),
		'yesterday': reverse('area_access') + '?' + urlencode({'start': yesterday, 'end': yesterday}),
	}
	try:
		start, end = parse_start_and_end_date(request.GET['start'], request.GET['end'])
		area_id = request.GET.get('area')
		dictionary['start'] = start
		dictionary['end'] = end
		area_access_records = AreaAccessRecord.objects.filter(start__gte=start, start__lt=end, staff_charge=None)
		if area_id:
			area_id = int(area_id)
			filter_areas = Area.objects.get(pk=area_id).get_descendants(include_self=True)
			area_access_records = area_access_records.filter(area__in=filter_areas)
			dictionary['area_id'] = area_id
		area_access_records = area_access_records.order_by('area__name')
		area_access_records.query.add_ordering(F('end').desc(nulls_first=True))
		area_access_records.query.add_ordering(F('start').desc())
		dictionary['access_records'] = area_access_records
	except:
		pass
	dictionary['area_select_field'] = TreeNodeChoiceField(Area.objects.filter(area_children_set__isnull=True).only('name'), empty_label="All").widget.render('area', area_id)
	return render(request, 'area_access/area_access.html', dictionary)


@staff_member_required
@require_http_methods(['GET', 'POST'])
def new_area_access_record(request):
	dictionary = {
		'customers': User.objects.filter(is_active=True)
	}
	if request.method == 'GET':
		try:
			customer = User.objects.get(id=request.GET['customer'])
			dictionary['customer'] = customer
			error_message = check_policy_for_user(customer=customer)
			if error_message:
				dictionary['error_message'] = error_message
				return render(request, 'area_access/new_area_access_record.html', dictionary)

			dictionary['user_accessible_areas'], dictionary['areas'] = load_areas_for_use_in_template(customer)
			return render(request, 'area_access/new_area_access_record_details.html', dictionary)
		except:
			pass
		return render(request, 'area_access/new_area_access_record.html', dictionary)
	if request.method == 'POST':
		try:
			user = User.objects.get(id=request.POST['customer'])
			project = Project.objects.get(id=request.POST['project'])
			area = Area.objects.get(id=request.POST['area'])
		except:
			dictionary['error_message'] = 'Your request contained an invalid identifier.'
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		try:
			error_message = check_policy_for_user(customer=user)
			if error_message:
				dictionary['error_message'] = error_message
				return render(request, 'area_access/new_area_access_record.html', dictionary)
			check_billing_to_project(project, user, area)
			check_policy_to_enter_this_area(area=area, user=user)
		except ProjectChargeException as e:
			dictionary['error_message'] = e.msg
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		except NoAccessiblePhysicalAccessUserError as error:
			if error.access_exception:
				dictionary['error_message'] = '{} does not have access to the {} at this time due to the following exception: {}.'.format(user, area.name, error.access_exception.name)
			else:
				dictionary['error_message'] = '{} does not have a physical access level that allows access to the {} at this time.'.format(user, area.name)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		except UnavailableResourcesUserError as error:
			dictionary['error_message'] = 'The {} is inaccessible because a required resource ({}) is unavailable. You must make all required resources for this area available before creating a new area access record.'.format(error.area.name, error.resources[0])
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		except MaximumCapacityReachedError as error:
			dictionary['error_message'] = 'The {} is inaccessible because the {} has reached its maximum capacity. Wait for somebody to exit and try again.'.format(area.name, error.area.name)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		except ScheduledOutageInProgressError as error:
			dictionary['error_message'] = 'The {} is inaccessible because a scheduled outage is in effect. You must wait for the outage to end before creating a new area access record.'.format(error.area.name)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		except ReservationRequiredUserError:
			dictionary['error_message'] = 'You do not have a current reservation for the {}. Please make a reservation before trying to access this area.'.format(area.name)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if user.billing_to_project():
			dictionary['error_message'] = '{} is already billing area access to another area. The user must log out of that area before entering another.'.format(user)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		if project not in user.active_projects():
			dictionary['error_message'] = '{} is not authorized to bill that project.'.format(user)
			return render(request, 'area_access/new_area_access_record.html', dictionary)
		record = AreaAccessRecord()
		record.area = area
		record.customer = user
		record.project = project
		record.save()
		dictionary['success'] = '{} is now logged in to the {}.'.format(user, area.name)
		return render(request, 'area_access/new_area_access_record.html', dictionary)


def check_policy_for_user(customer: User):
	error_message = None
	try:
		check_policy_to_enter_any_area(user=customer)
	except InactiveUserError:
		error_message = '{} is inactive'.format(customer)
	except NoActiveProjectsForUserError:
		error_message = '{} does not have any active projects to bill area access'.format(customer)
	except NoPhysicalAccessUserError:
		error_message = '{} does not have access to any billable areas'.format(customer)
	except PhysicalAccessExpiredUserError:
		error_message = '{} does not have access to any areas because the user\'s physical access expired on {}. You must update the user\'s physical access expiration date before creating a new area access record.'.format(customer, customer.access_expiration.strftime('%B %m, %Y'))
	return error_message


@staff_member_required
@require_POST
def force_area_logout(request, user_id):
	user = get_object_or_404(User, id=user_id)
	record = user.area_access_record()
	if record is None:
		return HttpResponseBadRequest('That user is not logged into any areas.')
	log_out_user(user)
	return HttpResponse()


@login_required
@require_http_methods(['GET', 'POST'])
def change_project(request, new_project=None):
	user: User = request.user
	""" For area access, allow the user to stop billing a project and start billing another project. """
	if request.method == 'GET':
		return render(request, 'area_access/change_project.html')
	old_project = user.billing_to_project()
	if old_project is None:
		dictionary = {
			'error': "There was a problem changing the project you're billing for area access. You must already be billing a project in order to change to another project, however you were not logged in to an area."
		}
		return render(request, 'area_access/change_project.html', dictionary)
	# If we're already billing the requested project then there's nothing to do.
	if old_project.id == new_project:
		return redirect(reverse('landing'))
	new_project = get_object_or_404(Project, id=new_project)
	try:
		check_billing_to_project(new_project, user, user.area_access_record().area)
	except ProjectChargeException as e:
		dictionary = {
			'error': e.msg
		}
		return render(request, 'area_access/change_project.html', dictionary)
	# Stop billing the user's initial project
	record = user.area_access_record()
	record.end = timezone.now()
	record.save()
	area = record.area
	# Start billing the user's new project
	record = AreaAccessRecord()
	record.area = area
	record.customer = request.user
	record.project = new_project
	record.save()
	return redirect(reverse('landing'))


@login_required
@require_POST
def calendar_self_login(request):
	"""
	This method is strongly dependent on the way self_log_in() works
	It looks for a redirect to landing as success, and an error message in a <div class="alert alert-danger"></div> for failure
	"""
	dictionary = {
		'projects': request.user.active_projects(),
	}
	try:
		a = Area.objects.get(id=request.POST['area'])
		dictionary['area'] = a
		Project.objects.get(id=request.POST.get('project'))
	except Project.DoesNotExist:
		# We have not selected a project yet
		return render(request, 'area_access/calendar_self_login.html', dictionary)
	response = self_log_in(request=request, load_areas=False)
	if response.status_code == 302 and response.url == reverse('landing'):
		# We got redirect to landing page in return, which means it was successful
		return HttpResponse()
	elif response and response.content:
		parser = ParseSelfLoginErrorMessage()
		parser.feed(response.content.decode())
		dictionary['error_message'] = parser.data
		return render(request, 'area_access/calendar_self_login.html', dictionary)


@login_required
@require_http_methods(['GET', 'POST'])
def self_log_in(request, load_areas=True):
	user: User = request.user
	if not able_to_self_log_in_to_area(user):
		return redirect(reverse('landing'))

	dictionary = {
		'projects': user.active_projects(),
	}
	if request.GET.get('area_id'):
		dictionary['area_id'] = quiet_int(request.GET['area_id'])

	facility_name = get_customization('facility_name')
	try:
		check_policy_to_enter_any_area(user)
	except InactiveUserError:
		dictionary['error_message'] = f'Your account has been deactivated. Please visit the {facility_name} staff to resolve the problem.'
		return render(request, 'area_access/self_login.html', dictionary)
	except NoActiveProjectsForUserError:
		dictionary['error_message'] = f"You are not a member of any active projects. You won't be able to use any interlocked {facility_name} tools. Please visit the {facility_name} user office for more information."
		return render(request, 'area_access/self_login.html', dictionary)
	except PhysicalAccessExpiredUserError:
		dictionary['error_message'] = f"Your physical access to the {facility_name} has expired. Have you completed your safety training within the last year? Please visit the User Office to renew your access."
		return render(request, 'area_access/self_login.html', dictionary)
	except NoPhysicalAccessUserError:
		dictionary['error_message'] = f"You have not been granted physical access to any {facility_name} area. Please visit the User Office if you believe this is an error."
		return render(request, 'area_access/self_login.html', dictionary)

	if load_areas:
		dictionary['user_accessible_areas'], dictionary['areas'] = load_areas_for_use_in_template(user)
	else:
		dictionary['user_accessible_areas'] = []
		dictionary['areas'] = []

	if request.method == 'GET':
		return render(request, 'area_access/self_login.html', dictionary)
	if request.method == 'POST':
		try:
			a = Area.objects.get(id=request.POST['area'])
			p = Project.objects.get(id=request.POST['project'])
			check_policy_to_enter_this_area(a, request.user)
			check_billing_to_project(p, user, a)
			log_in_user_to_area(a, request.user, p)
		except ProjectChargeException as e:
			dictionary['area_error_message'] = e.msg
			return render(request, 'area_access/self_login.html', dictionary)
		except NoAccessiblePhysicalAccessUserError as error:
			if error.access_exception:
				dictionary['area_error_message'] = f"You do not have access to the {error.area.name} at this time due to the following exception: {error.access_exception.name}. The exception ends on {localize(error.access_exception.end_time.astimezone(timezone.get_current_timezone()))}"
			else:
				dictionary['area_error_message'] = f"You do not have access to the {error.area.name} at this time. Please visit the User Office if you believe this is an error."
			return render(request, 'area_access/self_login.html', dictionary)
		except UnavailableResourcesUserError as error:
			dictionary['area_error_message'] = f'The {error.area.name} is inaccessible because a required resource is unavailable ({error.resources[0]}).'
			return render(request, 'area_access/self_login.html', dictionary)
		except ScheduledOutageInProgressError as error:
			dictionary['area_error_message'] = f'The {error.area.name} is inaccessible because a scheduled outage is in progress.'
			return render(request, 'area_access/self_login.html', dictionary)
		except MaximumCapacityReachedError as error:
			dictionary['area_error_message'] = f'The {error.area.name} is inaccessible because it has reached its maximum capacity. Wait for somebody to exit and try again.'
			return render(request, 'area_access/self_login.html', dictionary)
		except ReservationRequiredUserError as error:
			dictionary['area_error_message'] = f'You do not have a current reservation for the {error.area.name}. Please make a reservation before trying to access this area.'
			return render(request, 'area_access/self_login.html', dictionary)
		except Exception as error:
			area_access_logger.exception(error)
			dictionary['area_error_message'] = "unexpected error"
			return render(request, 'area_access/self_login.html', dictionary)
		return redirect(reverse('landing'))


@login_required
@require_GET
def self_log_out(request, user_id):
	user = get_object_or_404(User, id=user_id)
	if able_to_self_log_out_of_area(user):
		record = user.area_access_record()
		if record is None:
			return HttpResponseBadRequest('You are not logged into any areas.')
		log_out_user(user)
	return redirect(reverse('landing'))


@login_required
@require_GET
@disable_session_expiry_refresh
def occupancy(request):
	area_name = request.GET.get('occupancy')
	if area_name is None:
		return HttpResponse()
	try:
		area = Area.objects.get(name=area_name)
	except Area.DoesNotExist:
		return HttpResponse()
	dictionary = {
		'area': area,
		'occupants': AreaAccessRecord.objects.filter(area__name=area.name, end=None, staff_charge=None).prefetch_related('customer').order_by('-start'),
	}
	return render(request, 'occupancy/occupancy.html', dictionary)


@synchronized("user")
def log_in_user_to_area(area, user, project):
	return AreaAccessRecord.objects.create(area=area, customer=user, project=project)


@synchronized("user")
def log_out_user(user: User):
	record = user.area_access_record()
	# Allow the user to log out of any area, even if this is a logout tablet for a different area.
	if record:
		record.end = timezone.now()
		record.save()
		# Shorten the user's area reservation since the user is now leaving
		shorten_reservation(user, record.area)
		# Stop charging area access if staff is leaving the area
		staff_charge = user.get_staff_charge()
		if staff_charge:
			try:
				staff_area_access = AreaAccessRecord.objects.get(staff_charge=staff_charge, end=None)
				staff_area_access.end = timezone.now()
				staff_area_access.save()
			except AreaAccessRecord.DoesNotExist:
				pass


def able_to_self_log_out_of_area(user):
	# 'Self log out' must be enabled
	if not get_customization('self_log_out') == 'enabled':
		return False
	# Check if the user is active
	if not user.is_active:
		return False
	# Make sure the user is already in an area.
	if not user.in_area():
		return False
	# Otherwise we are good to log out
	return True


def able_to_self_log_in_to_area(user):
	# 'Self log in' must be enabled
	if not get_customization('self_log_in') == 'enabled':
		return False
	# Check if the user is already in an area. If so, the /change_project/ URL can be used to change their project.
	if user.in_area():
		return False

	# Otherwise user can try to self log in
	return True


def load_areas_for_use_in_template(user: User):
	"""
	This method returns accessible areas for the user and a queryset ready to be used in template view.
	The template view needs to use the {% recursetree %} tag from mptt
	"""
	accessible_areas = user.accessible_areas()
	areas = list(set([ancestor for area in accessible_areas for ancestor in area.get_ancestors(include_self=True)]))
	areas.sort(key=lambda x: x.tree_category())
	areas = Area.objects.filter(id__in=[area.id for area in areas])
	return accessible_areas, areas



from django.core.paginator import Paginator
from django.db.models import QuerySet


class SortedPaginator(Paginator):
    def __init__(self, object_list, request, per_page="25", order_by=None, orphans=0, allow_empty_first_page=True):
        per_page = request.GET.get("pp", per_page)
        self.page_number = request.GET.get("p")
        self.order_by = request.GET.get("o", order_by)
        if object_list and isinstance(object_list, QuerySet) and self.order_by:
            object_list = object_list.order_by(self.order_by)
        super().__init__(object_list, per_page, orphans, allow_empty_first_page)

    def get_current_page(self):
        return super().get_page(self.page_number)



from django_filters import filterset
from django_filters.rest_framework import DjangoFilterBackend
from django_filters.utils import try_dbfield


class NEMOFilterSet(filterset.FilterSet):
	""" This filter set for django rest_frameworks adds support for isempty lookup in GUI """

	@classmethod
	def filter_for_lookup(cls, field, lookup_type):
		DEFAULTS = dict(cls.FILTER_DEFAULTS)
		if hasattr(cls, "_meta"):
			DEFAULTS.update(cls._meta.filter_overrides)

		data = try_dbfield(DEFAULTS.get, field.__class__) or {}
		filter_class = data.get("filter_class")

		# if there is no filter class, exit early
		if not filter_class:
			return None, {}

		if lookup_type == "isempty":
			from django.db import models

			data = try_dbfield(DEFAULTS.get, models.BooleanField)

			filter_class = data.get("filter_class")
			params = data.get("extra", lambda x: {})(field)
			return filter_class, params
		else:
			return super().filter_for_lookup(field, lookup_type)


class NEMOFilterBackend(DjangoFilterBackend):
	filterset_base = NEMOFilterSet



# Generated by Django 2.2.10 on 2020-02-13 14:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0011_version_1_22_0'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tool',
            name='_post_usage_questions',
            field=models.TextField(blank=True, db_column='post_usage_questions', help_text='Upon logging off a tool, questions can be asked such as how much consumables were used by the user. This field will only accept JSON format', null=True),
        ),
    ]



from datetime import timedelta, datetime
from typing import Type, List

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from django.utils import timezone

from NEMO.models import News, Notification, SafetyIssue, User, BuddyRequest, BuddyRequestMessage
from NEMO.utilities import end_of_the_day


def delete_expired_notifications():
	Notification.objects.filter(expiration__lt=timezone.now()).delete()


def get_notifications(user: User, notification_type: Type[Model]):
	content_type = ContentType.objects.get_for_model(notification_type)
	notifications = Notification.objects.filter(user=user, content_type=content_type)
	if notifications:
		notification_ids = list(notifications.values_list("object_id", flat=True))
		notifications.delete()
		return notification_ids
	else:
		return None


def get_notification_counts(user: User):
	counts = {}
	for t in Notification.Types.Choices:
		model = t[0]
		content_type = ContentType.objects.get(app_label="NEMO", model=model)
		counts[model] = Notification.objects.filter(user=user, content_type=content_type).count()
	return counts


def create_news_notification(story):
	content_type = ContentType.objects.get_for_model(News)
	# Delete all existing notifications for this story, so we don't have multiple notifications for the same story
	Notification.objects.filter(
		content_type=content_type, object_id=story.id
	).delete()
	users = User.objects.filter(is_active=True)
	expiration = timezone.now() + timedelta(days=30)  # Unread news story notifications always expire after 30 days
	for u in users:
		Notification.objects.create(user=u, expiration=expiration, content_object=story)


def delete_news_notification(story):
	content_type = ContentType.objects.get_for_model(News)
	Notification.objects.filter(content_type=content_type, object_id=story.id).delete()


def create_safety_notification(safety_issue):
	users = User.objects.filter(is_staff=True, is_active=True)
	expiration = timezone.now() + timedelta(days=30)  # Unread safety issue notifications always expire after 30 days
	for u in users:
		Notification.objects.create(user=u, expiration=expiration, content_object=safety_issue)


def delete_safety_notification(issue):
	content_type = ContentType.objects.get_for_model(SafetyIssue)
	Notification.objects.filter(content_type=content_type, object_id=issue.id).delete()


def create_buddy_request_notification(buddy_request: BuddyRequest):
	users: List[User] = User.objects.filter(is_active=True).exclude(id=buddy_request.user_id)
	request_end = buddy_request.end
	# Unread buddy request notifications expire after the request ends
	expiration = end_of_the_day(
		datetime(request_end.year, request_end.month, request_end.day)
	)
	for u in users:
		if u.get_preferences().display_new_buddy_request_notification:
			Notification.objects.create(user=u, expiration=expiration, content_object=buddy_request)


def delete_buddy_request_notification(buddy_request: BuddyRequest):
	content_type = ContentType.objects.get_for_model(BuddyRequest)
	Notification.objects.filter(content_type=content_type, object_id=buddy_request.id).delete()


def create_buddy_reply_notification(reply: BuddyRequestMessage):
	creator: User = reply.buddy_request.user
	request_end = reply.buddy_request.end
	# Unread buddy request reply notifications expire after the request ends
	expiration = end_of_the_day(
		datetime(request_end.year, request_end.month, request_end.day)
	)
	for user in reply.buddy_request.creator_and_reply_users():
		if user != reply.author and (
				user == creator or user.get_preferences().display_new_buddy_request_reply_notification
		):
			Notification.objects.create(user=user, expiration=expiration, content_object=reply)



import sys
from json import loads

from django import forms
from django.contrib import admin
from django.contrib.admin import register
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import Permission
from django.db.models import Q
from django.db.models.fields.files import FieldFile
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from mptt.admin import DraggableMPTTAdmin, MPTTAdminForm, TreeRelatedFieldListFilter

from NEMO.actions import (
	duplicate_tool_configuration,
	lock_selected_interlocks,
	rebuild_area_tree,
	synchronize_with_tool_usage,
	unlock_selected_interlocks,
)
from NEMO.forms import BuddyRequestForm
from NEMO.models import (
	Account,
	AccountType,
	ActivityHistory,
	Alert,
	AlertCategory,
	Area,
	AreaAccessRecord,
	BadgeReader,
	BuddyRequest,
	BuddyRequestMessage,
	Comment,
	Configuration,
	ConfigurationHistory,
	Consumable,
	ConsumableCategory,
	ConsumableWithdraw,
	ContactInformation,
	ContactInformationCategory,
	Customization,
	Door,
	EmailLog,
	Interlock,
	InterlockCard,
	InterlockCardCategory,
	LandingPageChoice,
	MembershipHistory,
	News,
	Notification,
	PhysicalAccessException,
	PhysicalAccessLevel,
	PhysicalAccessLog,
	Project,
	Reservation,
	ReservationQuestions,
	Resource,
	ResourceCategory,
	SafetyIssue,
	ScheduledOutage,
	ScheduledOutageCategory,
	StaffCharge,
	Task,
	TaskCategory,
	TaskHistory,
	TaskImages,
	TaskStatus,
	Tool,
	ToolDocuments,
	ToolUsageCounter,
	TrainingSession,
	UsageEvent,
	User,
	UserPreferences,
	UserType,
	record_active_state,
	record_local_many_to_many_changes,
	record_remote_many_to_many_changes_and_save,
)
from NEMO.widgets.dynamic_form import (
	DynamicForm,
	PostUsageFloatFieldQuestion,
	PostUsageNumberFieldQuestion,
)


class ToolAdminForm(forms.ModelForm):
	class Meta:
		model = Tool
		fields = "__all__"

	class Media:
		js = ("admin/tool/tool.js", "admin/questions_preview/questions_preview.js")
		css = {"": ("admin/questions_preview/questions_preview.css",)}

	qualified_users = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Users", is_stacked=False),
	)

	required_resources = forms.ModelMultipleChoiceField(
		queryset=Resource.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Required resources", is_stacked=False),
	)

	nonrequired_resources = forms.ModelMultipleChoiceField(
		queryset=Resource.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Nonrequired resources", is_stacked=False),
	)

	_tool_calendar_color = forms.CharField(
		required=False, max_length=9, initial="#33ad33", widget=forms.TextInput(attrs={"type": "color"})
	)

	def __init__(self, *args, **kwargs):
		super(ToolAdminForm, self).__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields["qualified_users"].initial = self.instance.user_set.all()
			self.fields["required_resources"].initial = self.instance.required_resource_set.all()
			self.fields["nonrequired_resources"].initial = self.instance.nonrequired_resource_set.all()

	def clean(self):
		cleaned_data = super().clean()
		parent_tool = cleaned_data.get("parent_tool")
		category = cleaned_data.get("_category")
		location = cleaned_data.get("_location")
		phone_number = cleaned_data.get("_phone_number")
		primary_owner = cleaned_data.get("_primary_owner")
		image = cleaned_data.get("_image")

		# only resize if an image is present and  has changed
		if image and not isinstance(image, FieldFile):
			from NEMO.utilities import resize_image

			# resize image to 500x500 maximum
			cleaned_data["_image"] = resize_image(image, 500)

		if parent_tool:
			if parent_tool.id == self.instance.id:
				self.add_error("parent_tool", "You cannot select the parent to be the tool itself.")
			# in case of alternate tool, remove everything except parent_tool and name
			data = dict([(k, v) for k, v in self.cleaned_data.items() if k == "parent_tool" or k == "name"])
			# an alternate tool is never visible
			data["visible"] = False
			return data
		else:
			if not category:
				self.add_error("_category", "This field is required.")
			if not location:
				self.add_error("_location", "This field is required.")
			if not phone_number:
				self.add_error("_phone_number", "This field is required.")
			if not primary_owner:
				self.add_error("_primary_owner", "This field is required.")

			post_usage_questions = cleaned_data.get("_post_usage_questions")
			# Validate _post_usage_questions JSON format
			if post_usage_questions:
				try:
					loads(post_usage_questions)
				except ValueError:
					self.add_error("_post_usage_questions", "This field needs to be a valid JSON string")
				try:
					DynamicForm(post_usage_questions).validate("tool_usage_group_question", self.instance.id)
				except Exception:
					error_info = sys.exc_info()
					self.add_error("_post_usage_questions", error_info[0].__name__ + ": " + str(error_info[1]))

			policy_off_between_times = cleaned_data.get("_policy_off_between_times")
			policy_off_start_time = cleaned_data.get("_policy_off_start_time")
			policy_off_end_time = cleaned_data.get("_policy_off_end_time")
			if policy_off_between_times and (not policy_off_start_time or not policy_off_end_time):
				if not policy_off_start_time:
					self.add_error("_policy_off_start_time", "Start time must be specified")
				if not policy_off_end_time:
					self.add_error("_policy_off_end_time", "End time must be specified")


class ToolDocumentsInline(admin.TabularInline):
	model = ToolDocuments
	extra = 1


@register(Tool)
class ToolAdmin(admin.ModelAdmin):
	inlines = [ToolDocumentsInline]
	list_display = (
		"name_display",
		"_category",
		"visible",
		"operational_display",
		"problematic",
		"is_configurable",
		"id",
	)
	filter_horizontal = ("_backup_owners", "_superusers")
	search_fields = ("name", "_description", "_serial")
	list_filter = ("visible", "_operational", "_category", "_location")
	readonly_fields = ("_post_usage_preview",)
	actions = [duplicate_tool_configuration]
	form = ToolAdminForm
	fieldsets = (
		(
			None,
			{
				"fields": (
					"name",
					"parent_tool",
					"_category",
					"qualified_users",
					"_post_usage_questions",
					"_post_usage_preview",
				)
			},
		),
		("Additional Information", {"fields": ("_description", "_serial", "_image", "_tool_calendar_color")}),
		("Current state", {"fields": ("visible", "_operational")}),
		(
			"Contact information",
			{
				"fields": (
					"_primary_owner",
					"_backup_owners",
					"_superusers",
					"_notification_email_address",
					"_location",
					"_phone_number",
				)
			},
		),
		("Reservation", {"fields": ("_reservation_horizon", "_missed_reservation_threshold")}),
		(
			"Usage policy",
			{
				"fields": (
					"_policy_off_between_times",
					"_policy_off_start_time",
					"_policy_off_end_time",
					"_policy_off_weekend",
					"_minimum_usage_block_time",
					"_maximum_usage_block_time",
					"_maximum_reservations_per_day",
					"_minimum_time_between_reservations",
					"_maximum_future_reservation_time",
				)
			},
		),
		(
			"Area Access",
			{
				"fields": (
					"_requires_area_access",
					"_grant_physical_access_level_upon_qualification",
					"_grant_badge_reader_access_upon_qualification",
					"_interlock",
					"_allow_delayed_logoff",
				)
			},
		),
		("Dependencies", {"fields": ("required_resources", "nonrequired_resources")}),
	)

	def _post_usage_preview(self, obj):
		if obj.id:
			form_validity_div = '<div id="form_validity"></div>' if obj.post_usage_questions else ""
			return mark_safe(
				'<div class="questions_preview">{}{}</div><div class="help questions_preview_help">Save form to preview post usage questions</div>'.format(
					DynamicForm(obj.post_usage_questions).render("tool_usage_group_question", obj.id), form_validity_div
				)
			)

	def formfield_for_foreignkey(self, db_field, request, **kwargs):
		""" We only want non children tool to be eligible as parents """
		if db_field.name == "parent_tool":
			kwargs["queryset"] = Tool.objects.filter(parent_tool__isnull=True)
		return super().formfield_for_foreignkey(db_field, request, **kwargs)

	def save_model(self, request, obj, form, change):
		"""
		Explicitly record any project membership changes.
		"""
		if obj.parent_tool:
			if obj.pk:
				# if this is an update (from regular to child tool), we want to make sure we are creating a clean version. In case the previous tool had fields that are now irrelevant
				clean_alt_tool = Tool(**form.cleaned_data)
				clean_alt_tool.pk = obj.pk
				obj = clean_alt_tool
			super(ToolAdmin, self).save_model(request, obj, form, change)
		else:
			record_remote_many_to_many_changes_and_save(
				request, obj, form, change, "qualified_users", super(ToolAdmin, self).save_model
			)
			if "required_resources" in form.changed_data:
				obj.required_resource_set.set(form.cleaned_data["required_resources"])
			if "nonrequired_resources" in form.changed_data:
				obj.nonrequired_resource_set.set(form.cleaned_data["nonrequired_resources"])


class AreaAdminForm(MPTTAdminForm):
	class Meta:
		model = Area
		fields = "__all__"

	area_calendar_color = forms.CharField(
		required=False, max_length=9, initial="#88B7CD", widget=forms.TextInput(attrs={"type": "color"})
	)


@register(Area)
class AreaAdmin(DraggableMPTTAdmin):
	list_display = (
		"tree_actions",
		"indented_title",
		"name",
		"parent_area",
		"category",
		"requires_reservation",
		"maximum_capacity",
		"reservation_warning",
		"id",
	)
	fieldsets = (
		(None, {"fields": ("name", "parent_area", "category", "reservation_email", "abuse_email")}),
		("Additional Information", {"fields": ("area_calendar_color",)}),
		(
			"Area access",
			{"fields": ("requires_reservation", "logout_grace_period", "welcome_message", "buddy_system_allowed")},
		),
		(
			"Occupancy",
			{
				"fields": (
					"maximum_capacity",
					"count_staff_in_occupancy",
					"count_service_personnel_in_occupancy",
					"reservation_warning",
				)
			},
		),
		("Reservation", {"fields": ("reservation_horizon", "missed_reservation_threshold")}),
		(
			"Policy",
			{
				"fields": (
					"policy_off_between_times",
					"policy_off_start_time",
					"policy_off_end_time",
					"policy_off_weekend",
					"minimum_usage_block_time",
					"maximum_usage_block_time",
					"maximum_reservations_per_day",
					"minimum_time_between_reservations",
					"maximum_future_reservation_time",
				)
			},
		),
	)
	list_display_links = ("indented_title",)
	list_filter = ("requires_reservation", ("parent_area", TreeRelatedFieldListFilter))
	search_fields = ("name",)
	actions = [rebuild_area_tree]

	mptt_level_indent = 20
	form = AreaAdminForm

	def get_fieldsets(self, request, obj: Area = None):
		"""
		Remove some fieldsets if this area is a parent
		"""
		if obj and not obj.is_leaf_node():
			return [i for i in self.fieldsets if i[0] not in ["Area access", "Reservation", "Policy"]]
		return super().get_fieldsets(request, obj)

	def save_model(self, request, obj: Area, form, change):
		parent_area: Area = obj.parent_area
		if parent_area:
			# if this area has a parent, that parent needs to be cleaned and updated
			parent_area.is_now_a_parent()
		super(AreaAdmin, self).save_model(request, obj, form, change)


@register(TrainingSession)
class TrainingSessionAdmin(admin.ModelAdmin):
	list_display = ("id", "trainer", "trainee", "tool", "project", "type", "date", "duration", "qualified")
	list_filter = ("qualified", "date", "type", "tool")
	date_hierarchy = "date"

	def formfield_for_foreignkey(self, db_field, request, **kwargs):
		""" We only want staff user and tool superusers to be possible trainers """
		if db_field.name == "trainer":
			kwargs["queryset"] = User.objects.filter(Q(is_staff=True) | Q(superuser_for_tools__isnull=False)).distinct()
		return super().formfield_for_foreignkey(db_field, request, **kwargs)


@register(StaffCharge)
class StaffChargeAdmin(admin.ModelAdmin):
	list_display = ("id", "staff_member", "customer", "start", "end")
	list_filter = ("start",)
	date_hierarchy = "start"


@register(AreaAccessRecord)
class AreaAccessRecordAdmin(admin.ModelAdmin):
	list_display = ("id", "customer", "area", "project", "start", "end")
	list_filter = (("area", TreeRelatedFieldListFilter), "start")
	date_hierarchy = "start"


@register(Configuration)
class ConfigurationAdmin(admin.ModelAdmin):
	list_display = (
		"id",
		"tool",
		"name",
		"qualified_users_are_maintainers",
		"display_priority",
		"exclude_from_configuration_agenda",
	)
	filter_horizontal = ("maintainers",)


@register(ConfigurationHistory)
class ConfigurationHistoryAdmin(admin.ModelAdmin):
	list_display = ("id", "configuration", "user", "modification_time", "slot")
	date_hierarchy = "modification_time"


@register(Account)
class AccountAdmin(admin.ModelAdmin):
	list_display = ("name", "id", "active", "type", "start_date")
	search_fields = ("name",)
	list_filter = ("active", "type", "start_date")

	def save_model(self, request, obj, form, change):
		""" Audit account and project active status. """
		super(AccountAdmin, self).save_model(request, obj, form, change)
		record_active_state(request, obj, form, "active", not change)


class ProjectAdminForm(forms.ModelForm):
	class Meta:
		model = Project
		fields = "__all__"

	members = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Users", is_stacked=False),
	)

	principal_investigators = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Principal investigators", is_stacked=False),
	)

	def __init__(self, *args, **kwargs):
		super(ProjectAdminForm, self).__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields["members"].initial = self.instance.user_set.all()
			self.fields["principal_investigators"].initial = self.instance.manager_set.all()


@register(Project)
class ProjectAdmin(admin.ModelAdmin):
	fields = (
		"name",
		"application_identifier",
		"account",
		"start_date",
		"allow_consumable_withdrawals",
		"active",
		"members",
		"principal_investigators",
		"only_allow_tools",
	)
	list_display = ("name", "id", "application_identifier", "account", "active", "start_date")
	filter_horizontal = ("only_allow_tools",)
	search_fields = ("name", "application_identifier", "account__name")
	list_filter = ("active", "account", "start_date")
	form = ProjectAdminForm

	def save_model(self, request, obj, form, change):
		"""
		Audit project creation and modification. Also save any project membership changes explicitly.
		"""
		record_remote_many_to_many_changes_and_save(
			request, obj, form, change, "members", super(ProjectAdmin, self).save_model
		)
		# Make a history entry if a project has been moved under an account.
		# This applies to newly created projects and project ownership reassignment.
		if "account" in form.changed_data:
			# Create a membership removal entry for the project if it used to belong to another account:
			if change:
				previous_account = MembershipHistory()
				previous_account.authorizer = request.user
				previous_account.child_content_object = obj
				previous_account.parent_content_object = Account.objects.get(pk=form.initial["account"])
				previous_account.action = MembershipHistory.Action.REMOVED
				previous_account.save()

			# Create a membership addition entry for the project with its current account.
			current_account = MembershipHistory()
			current_account.authorizer = request.user
			current_account.child_content_object = obj
			current_account.parent_content_object = obj.account
			current_account.action = MembershipHistory.Action.ADDED
			current_account.save()

		# Record whether the project is active or not.
		record_active_state(request, obj, form, "active", not change)

		if "principal_investigators" in form.changed_data:
			obj.manager_set.set(form.cleaned_data["principal_investigators"])


@register(Reservation)
class ReservationAdmin(admin.ModelAdmin):
	list_display = (
		"id",
		"user",
		"creator",
		"tool",
		"area",
		"project",
		"start",
		"end",
		"duration",
		"cancelled",
		"missed",
	)
	readonly_fields = ("descendant",)
	list_filter = ("cancelled", "missed", "tool", "area")
	date_hierarchy = "start"


class ReservationQuestionsForm(forms.ModelForm):
	class Meta:
		model = ReservationQuestions
		fields = "__all__"

	class Media:
		js = ("admin/reservation_questions/reservation_questions.js", "admin/questions_preview/questions_preview.js")
		css = {"": ("admin/questions_preview/questions_preview.css",)}

	def clean(self):
		cleaned_data = super().clean()
		reservation_questions = cleaned_data.get("questions")
		tool_reservations = cleaned_data.get("tool_reservations")
		only_tools = cleaned_data.get("only_for_tools")
		area_reservations = cleaned_data.get("area_reservations")
		only_areas = cleaned_data.get("only_for_areas")
		if not tool_reservations and not area_reservations:
			self.add_error("tool_reservations", "Reservation questions have to apply to tool and/or area reservations")
			self.add_error("area_reservations", "Reservation questions have to apply to tool and/or area reservations")
		if not tool_reservations and only_tools:
			self.add_error(
				"tool_reservations", "You cannot restrict tools these questions apply to without enabling it for tools"
			)
		if not area_reservations and only_areas:
			self.add_error(
				"area_reservations", "You cannot restrict areas these questions apply to without enabling it for areas"
			)
		# Validate reservation_questions JSON format
		if reservation_questions:
			try:
				loads(reservation_questions)
			except ValueError:
				self.add_error("questions", "This field needs to be a valid JSON string")
			try:
				dynamic_form = DynamicForm(reservation_questions)
				dynamic_form.validate("reservation_group_question", self.instance.id)
			except KeyError as e:
				self.add_error("questions", f"{e} property is required")
			except Exception:
				error_info = sys.exc_info()
				self.add_error("questions", error_info[0].__name__ + ": " + str(error_info[1]))


@register(ReservationQuestions)
class ReservationQuestionsAdmin(admin.ModelAdmin):
	form = ReservationQuestionsForm
	filter_horizontal = ("only_for_tools", "only_for_areas", "only_for_projects")
	readonly_fields = ("questions_preview",)
	fieldsets = (
		(
			None,
			{
				"fields": (
					"name",
					"questions",
					"questions_preview",
					"tool_reservations",
					"only_for_tools",
					"area_reservations",
					"only_for_areas",
					"only_for_projects",
				)
			},
		),
	)

	def questions_preview(self, obj):
		form_validity_div = ""
		rendered_form = ""
		try:
			rendered_form = DynamicForm(obj.questions).render("reservation_group_question", obj.id)
			if obj.questions:
				form_validity_div = '<div id="form_validity"></div>'
		except:
			pass
		return mark_safe(
			'<div class="questions_preview">{}{}</div><div class="help questions_preview_help">Save form to preview reservation questions</div>'.format(
				rendered_form, form_validity_div
			)
		)


@register(UsageEvent)
class UsageEventAdmin(admin.ModelAdmin):
	list_display = ("id", "tool", "user", "operator", "project", "start", "end", "duration")
	list_filter = ("start", "end", "tool")
	date_hierarchy = "start"


@register(Consumable)
class ConsumableAdmin(admin.ModelAdmin):
	list_display = ("name", "quantity", "category", "visible", "reminder_threshold", "reminder_email", "id")
	list_filter = ("visible", "category")
	readonly_fields = ("reminder_threshold_reached",)


@register(ConsumableCategory)
class ConsumableCategoryAdmin(admin.ModelAdmin):
	list_display = ("name",)


@register(ConsumableWithdraw)
class ConsumableWithdrawAdmin(admin.ModelAdmin):
	list_display = ("id", "customer", "merchant", "consumable", "quantity", "project", "date")
	list_filter = ("date", "consumable")
	date_hierarchy = "date"


class InterlockCardAdminForm(forms.ModelForm):
	class Meta:
		model = InterlockCard
		widgets = {"password": forms.PasswordInput(render_value=True)}
		fields = "__all__"

	def clean(self):
		if any(self.errors):
			return
		super(InterlockCardAdminForm, self).clean()
		category = self.cleaned_data["category"]
		from NEMO import interlocks

		interlocks.get(category, False).clean_interlock_card(self)


@register(InterlockCard)
class InterlockCardAdmin(admin.ModelAdmin):
	form = InterlockCardAdminForm
	list_display = ("name", "server", "port", "number", "category", "even_port", "odd_port")


class InterlockAdminForm(forms.ModelForm):
	class Meta:
		model = Interlock
		fields = "__all__"

	def clean(self):
		if any(self.errors):
			return
		super(InterlockAdminForm, self).clean()
		from NEMO import interlocks

		category = self.cleaned_data["card"].category
		interlocks.get(category, False).clean_interlock(self)


@register(Interlock)
class InterlockAdmin(admin.ModelAdmin):
	form = InterlockAdminForm
	list_display = ("id", "card", "channel", "state", "tool", "door")
	actions = [lock_selected_interlocks, unlock_selected_interlocks, synchronize_with_tool_usage]
	readonly_fields = ["state", "most_recent_reply"]


@register(InterlockCardCategory)
class InterlockCardCategoryAdmin(admin.ModelAdmin):
	list_display = ("name",)


@register(Task)
class TaskAdmin(admin.ModelAdmin):
	list_display = (
		"id",
		"urgency",
		"tool",
		"creator",
		"creation_time",
		"last_updated",
		"problem_category",
		"cancelled",
		"resolved",
		"resolution_category",
	)
	list_filter = ("urgency", "resolved", "cancelled", "safety_hazard", "creation_time", "tool")
	date_hierarchy = "creation_time"


@register(TaskCategory)
class TaskCategoryAdmin(admin.ModelAdmin):
	list_display = ("name", "stage")


@register(TaskStatus)
class TaskStatusAdmin(admin.ModelAdmin):
	list_display = (
		"name",
		"notify_primary_tool_owner",
		"notify_backup_tool_owners",
		"notify_tool_notification_email",
		"custom_notification_email_address",
	)


@register(TaskHistory)
class TaskHistoryAdmin(admin.ModelAdmin):
	list_display = ("id", "task", "status", "time", "user")
	readonly_fields = ("time",)
	date_hierarchy = "time"


@register(TaskImages)
class TaskImagesAdmin(admin.ModelAdmin):
	list_display = ("id", "get_tool", "task", "uploaded_at")

	def get_tool(self, task_image: TaskImages):
		return task_image.task.tool.name

	get_tool.admin_order_field = "tool"  # Allows column order sorting
	get_tool.short_description = "Tool Name"  # Renames column head


@register(Comment)
class CommentAdmin(admin.ModelAdmin):
	list_display = (
		"id",
		"tool",
		"author",
		"creation_date",
		"expiration_date",
		"visible",
		"staff_only",
		"hidden_by",
		"hide_date",
	)
	list_filter = ("visible", "creation_date", "tool", "staff_only")
	date_hierarchy = "creation_date"
	search_fields = ("content",)


@register(Resource)
class ResourceAdmin(admin.ModelAdmin):
	list_display = ("name", "category", "available")
	list_filter = ("available", "category")
	filter_horizontal = ("fully_dependent_tools", "partially_dependent_tools", "dependent_areas")


@register(ActivityHistory)
class ActivityHistoryAdmin(admin.ModelAdmin):
	list_display = ("__str__", "content_type", "object_id", "action", "date", "authorizer")
	date_hierarchy = "date"


@register(MembershipHistory)
class MembershipHistoryAdmin(admin.ModelAdmin):
	list_display = (
		"__str__",
		"parent_content_type",
		"parent_object_id",
		"action",
		"child_content_type",
		"child_object_id",
		"date",
		"authorizer",
	)
	date_hierarchy = "date"


@register(UserType)
class UserTypeAdmin(admin.ModelAdmin):
	list_display = ("name",)


@register(UserPreferences)
class UserPreferencesAdmin(admin.ModelAdmin):
	list_display = ("user",)


class UserAdminForm(forms.ModelForm):
	class Meta:
		model = User
		fields = "__all__"

	backup_owner_on_tools = forms.ModelMultipleChoiceField(
		queryset=Tool.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="tools", is_stacked=False),
	)

	superuser_on_tools = forms.ModelMultipleChoiceField(
		queryset=Tool.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="tools", is_stacked=False),
	)

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields["backup_owner_on_tools"].initial = self.instance.backup_for_tools.all()
			self.fields["superuser_on_tools"].initial = self.instance.superuser_for_tools.all()

	def clean(self):
		cleaned_data = super().clean()
		staff_status = cleaned_data.get("is_staff")
		service_personnel_status = cleaned_data.get("is_service_personnel")

		if staff_status and service_personnel_status:
			raise forms.ValidationError(
				{
					"is_staff": "A user cannot be both staff and service personnel. Please choose one or the other.",
					"is_service_personnel": "A user cannot be both staff and service personnel. Please choose one or the other.",
				}
			)


@register(User)
class UserAdmin(admin.ModelAdmin):
	form = UserAdminForm
	filter_horizontal = (
		"groups",
		"user_permissions",
		"qualifications",
		"projects",
		"managed_projects",
		"physical_access_levels",
	)
	fieldsets = (
		(
			"Personal information",
			{"fields": ("first_name", "last_name", "username", "email", "badge_number", "type", "domain")},
		),
		(
			"Permissions",
			{
				"fields": (
					"is_active",
					"is_staff",
					"is_technician",
					"is_service_personnel",
					"is_superuser",
					"training_required",
					"groups",
					"user_permissions",
					"physical_access_levels",
				)
			},
		),
		("Important dates", {"fields": ("date_joined", "last_login", "access_expiration")}),
		(
			"Facility information",
			{
				"fields": (
					"qualifications",
					"backup_owner_on_tools",
					"superuser_on_tools",
					"projects",
					"managed_projects",
				)
			},
		),
	)
	search_fields = ("first_name", "last_name", "username", "email")
	list_display = (
		"first_name",
		"last_name",
		"username",
		"email",
		"is_active",
		"domain",
		"is_staff",
		"is_technician",
		"is_service_personnel",
		"is_superuser",
		"date_joined",
		"last_login",
	)
	list_filter = (
		"is_active",
		"domain",
		"is_staff",
		"is_technician",
		"is_service_personnel",
		"is_superuser",
		"date_joined",
		"last_login",
	)

	def formfield_for_manytomany(self, db_field, request, **kwargs):
		if db_field.name == "qualifications":
			kwargs["queryset"] = Tool.objects.filter(parent_tool__isnull=True)
		return super().formfield_for_manytomany(db_field, request, **kwargs)

	def save_model(self, request, obj, form, change):
		""" Audit project membership and qualifications when a user is saved. """
		super(UserAdmin, self).save_model(request, obj, form, change)
		record_local_many_to_many_changes(request, obj, form, "projects")
		record_local_many_to_many_changes(request, obj, form, "qualifications")
		record_local_many_to_many_changes(request, obj, form, "physical_access_levels")
		record_active_state(request, obj, form, "is_active", not change)
		if "backup_owner_on_tools" in form.changed_data:
			obj.backup_for_tools.set(form.cleaned_data["backup_owner_on_tools"])
		if "superuser_on_tools" in form.changed_data:
			obj.superuser_for_tools.set(form.cleaned_data["superuser_on_tools"])


@register(PhysicalAccessLog)
class PhysicalAccessLogAdmin(admin.ModelAdmin):
	list_display = ("user", "door", "time", "result")
	list_filter = ("door", "result")
	search_fields = ("user__first_name", "user__last_name", "user__username", "door__name")
	date_hierarchy = "time"

	def has_delete_permission(self, request, obj=None):
		return False

	def has_add_permission(self, request):
		return False

	def has_change_permission(self, request, obj=None):
		return False


@register(SafetyIssue)
class SafetyIssueAdmin(admin.ModelAdmin):
	list_display = ("id", "reporter", "creation_time", "visible", "resolved", "resolution_time", "resolver")
	list_filter = ("resolved", "visible", "creation_time", "resolution_time")
	readonly_fields = ("creation_time", "resolution_time")
	search_fields = ("location", "concern", "progress", "resolution")


@register(Door)
class DoorAdmin(admin.ModelAdmin):
	list_display = ("name", "area", "interlock", "get_absolute_url", "id")


@register(AlertCategory)
class AlertCategoryAdmin(admin.ModelAdmin):
	list_display = ("name",)


class AlertAdminForm(forms.ModelForm):
	contents = forms.CharField(widget=forms.Textarea(attrs={"rows": 3, "cols": 50}))

	class Meta:
		model = Alert
		fields = "__all__"


@register(Alert)
class AlertAdmin(admin.ModelAdmin):
	list_display = (
		"title",
		"category",
		"creation_time",
		"creator",
		"debut_time",
		"expiration_time",
		"user",
		"dismissible",
		"expired",
		"deleted",
	)
	form = AlertAdminForm


class PhysicalAccessLevelForm(forms.ModelForm):
	class Meta:
		model = PhysicalAccessLevel
		fields = "__all__"

	class Media:
		js = ("admin/physical_access_level/access_level.js",)

	authorized_users = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Users", is_stacked=False),
	)

	physical_access_exceptions = forms.ModelMultipleChoiceField(
		queryset=PhysicalAccessException.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(verbose_name="Physical Access Exceptions", is_stacked=False),
	)

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields["authorized_users"].initial = self.instance.user_set.all()
			self.fields["physical_access_exceptions"].initial = self.instance.physicalaccessexception_set.all()

	def clean(self):
		schedule = self.cleaned_data.get("schedule")
		if schedule == PhysicalAccessLevel.Schedule.WEEKDAYS:
			start_date = self.cleaned_data.get("weekdays_start_time")
			end_date = self.cleaned_data.get("weekdays_end_time")
			if not start_date:
				self.add_error("weekdays_start_time", "Start time is required for weekdays.")
			if not end_date:
				self.add_error("weekdays_end_time", "End time is required for weekdays.")
		else:
			self.cleaned_data["weekdays_start_time"] = None
			self.cleaned_data["weekdays_end_time"] = None
		return self.cleaned_data


@register(PhysicalAccessLevel)
class PhysicalAccessLevelAdmin(admin.ModelAdmin):
	form = PhysicalAccessLevelForm
	list_display = ("name", "area", "get_schedule_display_with_times", "allow_staff_access")
	list_filter = (("area", TreeRelatedFieldListFilter),)

	def save_model(self, request, obj, form, change):
		"""
		Explicitly record any membership changes.
		"""
		record_remote_many_to_many_changes_and_save(request, obj, form, change, "authorized_users", super().save_model)
		if "physical_access_exceptions" in form.changed_data:
			obj.physicalaccessexception_set.set(form.cleaned_data["physical_access_exceptions"])


class PhysicalAccessExceptionAdminForm(forms.ModelForm):
	class Meta:
		model = PhysicalAccessException
		fields = "__all__"

	class Media:
		js = ("admin/time_options_override.js",)

	def clean(self):
		if any(self.errors):
			return
		cleaned_data = super().clean()
		start_time = cleaned_data.get("start_time")
		end_time = cleaned_data.get("end_time")
		if end_time <= start_time:
			self.add_error("end_time", "The end time must be later than the start time")


@register(PhysicalAccessException)
class PhysicalAccessExceptionAdmin(admin.ModelAdmin):
	form = PhysicalAccessExceptionAdminForm
	list_display = ("name", "start_time", "end_time")
	filter_horizontal = ("physical_access_levels",)
	list_filter = ("physical_access_levels__area",)


@register(ContactInformationCategory)
class ContactInformationCategoryAdmin(admin.ModelAdmin):
	list_display = ("name", "display_order")


@register(ContactInformation)
class ContactInformationAdmin(admin.ModelAdmin):
	list_display = ("name", "category", "user")


@register(LandingPageChoice)
class LandingPageChoiceAdmin(admin.ModelAdmin):
	list_display = (
		"display_priority",
		"name",
		"url",
		"open_in_new_tab",
		"secure_referral",
		"hide_from_mobile_devices",
		"hide_from_desktop_computers",
	)
	list_display_links = ("name",)


@register(Customization)
class CustomizationAdmin(admin.ModelAdmin):
	list_display = ("name", "value")


@register(ScheduledOutageCategory)
class ScheduledOutageCategoryAdmin(admin.ModelAdmin):
	list_display = ("name",)


@register(ScheduledOutage)
class ScheduledOutageAdmin(admin.ModelAdmin):
	list_display = ("id", "tool", "area", "resource", "creator", "title", "start", "end")


@register(News)
class NewsAdmin(admin.ModelAdmin):
	list_display = ("id", "title", "created", "last_updated", "archived", "pinned")
	list_filter = ("archived", "pinned")


@register(Notification)
class NotificationAdmin(admin.ModelAdmin):
	list_display = ("id", "user", "expiration", "content_type", "object_id")


@register(BadgeReader)
class BadgeReaderAdmin(admin.ModelAdmin):
	list_display = ("id", "name", "send_key", "record_key")


class CounterAdminForm(forms.ModelForm):
	class Meta:
		model = ToolUsageCounter
		fields = "__all__"

	def clean(self):
		cleaned_data = super().clean()
		tool = cleaned_data.get("tool")
		tool_usage_question_name = cleaned_data.get("tool_usage_question")
		if tool and tool_usage_question_name:
			error = None
			if tool.post_usage_questions:
				post_usage_form = DynamicForm(tool.post_usage_questions)
				tool_question = post_usage_form.filter_questions(
					lambda x: (isinstance(x, PostUsageNumberFieldQuestion) or isinstance(x, PostUsageFloatFieldQuestion)) and x.name == tool_usage_question_name
				)
				if not tool_question:
					candidates = [
						question.name
						for question in post_usage_form.filter_questions(
							lambda x: isinstance(x, PostUsageNumberFieldQuestion) or isinstance(x, PostUsageFloatFieldQuestion)
						)
					]
					error = "The tool has no post usage question of type Number or Float with this name."
					if candidates:
						error += f" Valid question names are: {', '.join(candidates)}"
			else:
				error = "The tool does not have any post usage questions."
			if error:
				self.add_error("tool_usage_question", error)
		return cleaned_data


@register(ToolUsageCounter)
class CounterAdmin(admin.ModelAdmin):
	list_display = (
		"name",
		"tool",
		"tool_usage_question",
		"value",
		"warning_threshold",
		"last_reset",
		"last_reset_by",
		"is_active",
	)
	list_filter = ("tool",)
	readonly_fields = ("warning_threshold_reached",)
	form = CounterAdminForm


@register(BuddyRequest)
class BuddyRequestAdmin(admin.ModelAdmin):
	form = BuddyRequestForm
	list_display = ("user", "start", "end", "area", "reply_count", "expired", "deleted")
	list_filter = ("expired", "deleted")

	def reply_count(self, buddy_request: BuddyRequest):
		return buddy_request.replies.count()

	reply_count.admin_order_field = "replies"
	reply_count.short_description = "Replies"


@register(BuddyRequestMessage)
class BuddyRequestMessageAdmin(admin.ModelAdmin):
	list_display = ("id", "link_to_buddy_request", "author", "creation_date")

	def link_to_buddy_request(self, obj):
		link = reverse("admin:NEMO_buddyrequest_change", args=[obj.buddy_request.id])  # model name has to be lowercase
		return format_html('<a href="%s">%s</a>' % (link, obj.buddy_request))

	link_to_buddy_request.short_description = "BUDDY REQUEST"


@register(EmailLog)
class EmailLogAdmin(admin.ModelAdmin):
	list_display = ["id", "category", "sender", "to", "subject", "when", "ok"]
	list_filter = ["category", "ok"]
	search_fields = ["subject", "content", "to"]
	readonly_fields = ("content_preview",)
	date_hierarchy = "when"

	def content_preview(self, obj):
		return mark_safe(obj.content)

	def has_delete_permission(self, request, obj=None):
		return False

	def has_add_permission(self, request):
		return False

	def has_change_permission(self, request, obj=None):
		return False


admin.site.register(AccountType)
admin.site.register(ResourceCategory)
admin.site.register(Permission)



from datetime import datetime, timedelta
from itertools import chain

from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404, render
from django.utils import timezone
from django.utils.dateparse import parse_date, parse_time
from django.views.decorators.http import require_GET, require_POST

from NEMO.exceptions import ProjectChargeException, RequiredUnansweredQuestionsException
from NEMO.models import Area, Project, Reservation, ReservationItemType, ScheduledOutage, Tool, User
from NEMO.utilities import beginning_of_the_day, end_of_the_day, extract_date, localize
from NEMO.views.calendar import (
	determine_insufficient_notice,
	extract_configuration,
	extract_reservation_questions,
	render_reservation_questions,
)
from NEMO.views.customization import get_customization
from NEMO.views.policy import check_billing_to_project, check_policy_to_save_reservation


@login_required
@require_GET
def choose_item(request, next_page):
	user: User = request.user
	tools = Tool.objects.filter(visible=True).order_by('_category', 'name')
	dictionary = {
		'tools': list(tools),
		'areas': [],
	}
	if next_page == 'view_calendar':
		# If the user has no active projects then they're not allowed to make reservations. Redirect them home.
		if user.active_project_count() == 0:
			return render(request, 'mobile/no_active_projects.html')
		areas = Area.objects.filter(requires_reservation=True).only('name')
		# We want to remove areas the user doesn't have access to
		display_all_areas = get_customization('calendar_display_not_qualified_areas') == 'enabled'
		if not display_all_areas and areas and user and not user.is_superuser:
			areas = [area for area in areas if area in user.accessible_areas()]

		tool_area = 'tool/area'
		if tools and not areas:
			tool_area = 'tool'
		if areas and not tools:
			tool_area = 'area'

		dictionary['areas'] = list(areas)
		dictionary['title'] = f"Which {tool_area} calendar would you like to view?"
		dictionary['next_page'] = 'view_calendar'
	elif next_page == 'tool_control':
		dictionary['title'] = "Which tool control page would you like to view?"
		dictionary['next_page'] = 'tool_control'
	return render(request, 'mobile/choose_item.html', dictionary)


@login_required
@require_GET
def new_reservation(request, item_type, item_id, date=None):
	# If the user has no active projects then they're not allowed to make reservations.
	user: User = request.user
	if user.active_project_count() == 0:
		return render(request, 'mobile/no_active_projects.html')

	item_type = ReservationItemType(item_type)
	item = get_object_or_404(item_type.get_object_class(), id=item_id)
	if item_type == ReservationItemType.TOOL:
		dictionary = item.get_configuration_information(user=request.user, start=None)
	else:
		dictionary = {}
	dictionary['item'] = item
	dictionary['item_type'] = item_type.value
	dictionary['date'] = date
	dictionary['item_reservation_times'] = list(Reservation.objects.filter(**{item_type.value: item}).filter(cancelled=False, missed=False, shortened=False, start__gte=timezone.now()))

	# Reservation questions if applicable
	if not user.is_staff:
		reservation_question_dict = {}
		for project in user.active_projects():
			reservation_questions = render_reservation_questions(item_type, item_id, project)
			if reservation_questions:
				reservation_question_dict[project.id] = reservation_questions
		dictionary['reservation_questions'] = reservation_question_dict

	return render(request, 'mobile/new_reservation.html', dictionary)


@login_required
@require_POST
def make_reservation(request):
	""" Create a reservation for a user. """
	try:
		date = parse_date(request.POST['date'])
		start = localize(datetime.combine(date, parse_time(request.POST['start'])))
		end = localize(datetime.combine(date, parse_time(request.POST['end'])))
	except:
		return render(request, 'mobile/error.html', {'message': 'Please enter a valid date, start time, and end time for the reservation.'})
	item_type = ReservationItemType(request.POST['item_type'])
	item = get_object_or_404(item_type.get_object_class(), id=request.POST.get('item_id'))
	# Create the new reservation:
	reservation = Reservation()
	reservation.user = request.user
	reservation.creator = request.user
	reservation.reservation_item = item
	reservation.start = start
	reservation.end = end
	if item_type == ReservationItemType.TOOL:
		reservation.short_notice = determine_insufficient_notice(item, start)
	else:
		reservation.short_notice = False
	policy_problems, overridable = check_policy_to_save_reservation(cancelled_reservation=None, new_reservation=reservation, user_creating_reservation=request.user, explicit_policy_override=False)

	# If there was a problem in saving the reservation then return the error...
	if policy_problems:
		return render(request, 'mobile/error.html', {'message': policy_problems[0]})

	# All policy checks have passed.
	try:
		reservation.project = Project.objects.get(id=request.POST['project_id'])
		# Check if we are allowed to bill to project
		check_billing_to_project(reservation.project, request.user, reservation.reservation_item)
	except ProjectChargeException as e:
		return render(request, 'mobile/error.html', {'message': e.msg})
	except:
		if not request.user.is_staff:
			return render(request, 'mobile/error.html', {'message': 'You must specify a project for your reservation'})

	reservation.additional_information, reservation.self_configuration = extract_configuration(request)
	# Reservation can't be short notice if the user is configuring the tool themselves.
	if reservation.self_configuration:
		reservation.short_notice = False

	# Reservation questions if applicable
	try:
		reservation.question_data = extract_reservation_questions(request, item_type, item.id, reservation.project)
	except RequiredUnansweredQuestionsException as e:
		return render(request, 'mobile/error.html', {'message': str(e)})

	reservation.save_and_notify()
	return render(request, 'mobile/reservation_success.html', {'new_reservation': reservation})


@login_required
@require_GET
def view_calendar(request, item_type, item_id, date=None):
	item_type = ReservationItemType(item_type)
	item = get_object_or_404(item_type.get_object_class(), id=item_id)
	if date:
		try:
			date = extract_date(date)
		except:
			render(request, 'mobile/error.html', {'message': 'Invalid date requested for tool calendar'})
			return HttpResponseBadRequest()
	else:
		date = datetime.now()

	start = beginning_of_the_day(date, in_local_timezone=True)
	end = end_of_the_day(date, in_local_timezone=True)

	reservations = Reservation.objects.filter(**{item_type.value: item}).filter(cancelled=False, missed=False, shortened=False).filter(**{})
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	reservations = reservations.exclude(start__lt=start, end__lt=start)
	reservations = reservations.exclude(start__gt=end, end__gt=end)

	outages = ScheduledOutage.objects.none()
	if item_type == ReservationItemType.TOOL:
		outages = ScheduledOutage.objects.filter(Q(tool=item) | Q(resource__fully_dependent_tools__in=[item]))
	elif item_type == ReservationItemType.AREA:
		outages = item.scheduled_outage_queryset()

	# Exclude outages for which the following is true:
	# The outage starts and ends before the time-window, and...
	# The outage starts and ends after the time-window.
	outages = outages.exclude(start__lt=start, end__lt=start)
	outages = outages.exclude(start__gt=end, end__gt=end)

	events = list(chain(reservations, outages))
	events.sort(key=lambda x: x.start)

	dictionary = {
		'item': item,
		'item_type': item_type.value,
		'previous_day': start - timedelta(days=1),
		'current_day': start,
		'current_day_string': date.strftime('%Y-%m-%d'),
		'next_day': start + timedelta(days=1),
		'events': events,
	}

	return render(request, 'mobile/view_calendar.html', dictionary)



from itertools import chain

from django.db.models import Q
from django.shortcuts import get_object_or_404, render
from django.views.decorators.http import require_GET

from NEMO.decorators import staff_member_required
from NEMO.models import Task, TaskCategory, TaskStatus


@staff_member_required
@require_GET
def maintenance(request, sort_by=''):
	pending_tasks = Task.objects.filter(cancelled=False, resolved=False)
	if sort_by in ['urgency', 'force_shutdown', 'tool', 'problem_category', 'last_updated', 'creation_time']:
		if sort_by == 'last_updated':
			pending_tasks = pending_tasks.exclude(last_updated=None).order_by('-last_updated')
			not_yet_updated_tasks = Task.objects.filter(cancelled=False, resolved=False, last_updated=None).order_by('-creation_time')
			pending_tasks = list(chain(pending_tasks, not_yet_updated_tasks))
		else:
			pending_tasks = pending_tasks.order_by(sort_by)
			if sort_by in ['urgency', 'force_shutdown', 'creation_time']:
				pending_tasks = pending_tasks.reverse()
	else:
		pending_tasks = pending_tasks.order_by('urgency').reverse()  # Order by urgency by default
	closed_tasks = Task.objects.filter(Q(cancelled=True) | Q(resolved=True)).exclude(resolution_time__isnull=True).order_by('-resolution_time')[:20]
	dictionary = {
		'pending_tasks': pending_tasks,
		'closed_tasks': closed_tasks,
	}
	return render(request, 'maintenance/maintenance.html', dictionary)


@staff_member_required
@require_GET
def task_details(request, task_id):
	task = get_object_or_404(Task, id=task_id)

	if task.cancelled or task.resolved:
		return render(request, 'maintenance/closed_task_details.html', {'task': task})

	dictionary = {
		'task': task,
		'initial_assessment_categories': TaskCategory.objects.filter(stage=TaskCategory.Stage.INITIAL_ASSESSMENT),
		'completion_categories': TaskCategory.objects.filter(stage=TaskCategory.Stage.COMPLETION),
		'task_statuses': TaskStatus.objects.all(),
	}

	if task.tool.is_configurable():
		dictionary['rendered_configuration_html'] = task.tool.configuration_widget(request.user)

	return render(request, 'maintenance/pending_task_details.html', dictionary)



# Generated by Django 2.2.10 on 2020-06-03 19:50

import django.db.models.deletion
import mptt
import mptt.fields
import mptt.managers
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0019_user_type_not_required'),
    ]

    def rebuild_area_tree(apps, schema_editor):
        manager = mptt.managers.TreeManager()
        Area = apps.get_model("NEMO", "Area")
        manager.model = Area
        mptt.register(Area, parent_attr='parent_area')
        manager.contribute_to_class(Area, 'objects')
        manager.rebuild()

    operations = [
        migrations.AlterModelOptions(
            name='area',
            options={},
        ),
        migrations.AddField(
            model_name='area',
            name='level',
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='area',
            name='lft',
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='area',
            name='rght',
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='area',
            name='tree_id',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='area',
            name='category',
            field=models.CharField(blank=True, db_column='category', help_text='Create sub-categories using slashes. For example "Category 1/Sub-category 1".', max_length=1000, null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='parent_area',
            field=mptt.fields.TreeForeignKey(blank=True, help_text='Select a parent area, (building, floor etc.)', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='area_children_set', to='NEMO.Area'),
        ),
        migrations.AddField(
            model_name='area',
            name='logout_grace_period',
            field=models.PositiveIntegerField(blank=True, help_text='Number of minutes users have to logout of this area after their reservation expired before being flagged and abuse email is sent.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='maximum_future_reservation_time',
            field=models.PositiveIntegerField(blank=True, db_column='maximum_future_reservation_time', help_text='The maximum amount of time (in minutes) that a user may reserve from the current time onwards.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='maximum_reservations_per_day',
            field=models.PositiveIntegerField(blank=True, db_column='maximum_reservations_per_day', help_text='The maximum number of reservations a user may make per day for this area.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='maximum_usage_block_time',
            field=models.PositiveIntegerField(blank=True, db_column='maximum_usage_block_time', help_text='The maximum amount of time (in minutes) that a user may reserve this area for a single reservation. Leave this field blank to indicate that no maximum usage block time exists for this area.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='minimum_time_between_reservations',
            field=models.PositiveIntegerField(blank=True, db_column='minimum_time_between_reservations', help_text='The minimum amount of time (in minutes) that the same user must have between any two reservations for this area.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='minimum_usage_block_time',
            field=models.PositiveIntegerField(blank=True, db_column='minimum_usage_block_time', help_text='The minimum amount of time (in minutes) that a user must reserve this area for a single reservation. Leave this field blank to indicate that no minimum usage block time exists for this area.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='missed_reservation_threshold',
            field=models.PositiveIntegerField(blank=True, db_column='missed_reservation_threshold', help_text='The amount of time (in minutes) that a area reservation may go unused before it is automatically marked as "missed" and hidden from the calendar. Usage can be from any user, regardless of who the reservation was originally created for. The cancellation process is triggered by a timed job on the web server.', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='policy_off_between_times',
            field=models.BooleanField(db_column='policy_off_between_times', default=False, help_text='Check this box to disable policy rules every day between the given times'),
        ),
        migrations.AddField(
            model_name='area',
            name='policy_off_end_time',
            field=models.TimeField(blank=True, db_column='policy_off_end_time', help_text='The end time when policy rules should NOT be enforced', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='policy_off_start_time',
            field=models.TimeField(blank=True, db_column='policy_off_start_time', help_text='The start time when policy rules should NOT be enforced', null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='policy_off_weekend',
            field=models.BooleanField(db_column='policy_off_weekend', default=False, help_text='Whether or not policy rules should be enforced on weekends'),
        ),
        migrations.AddField(
            model_name='area',
            name='requires_reservation',
            field=models.BooleanField(default=False, help_text='Check this box to require a reservation for this area before a user can login.'),
        ),
        migrations.AddField(
            model_name='area',
            name='reservation_horizon',
            field=models.PositiveIntegerField(blank=True, db_column='reservation_horizon', default=14, help_text='Users may create reservations this many days in advance. Leave this field blank to indicate that no reservation horizon exists for this area.', null=True),
        ),
        migrations.AddField(
            model_name='scheduledoutage',
            name='area',
            field=mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area'),
        ),
        migrations.AddField(
            model_name='reservation',
            name='area',
            field=mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area'),
        ),
        migrations.AlterField(
            model_name='scheduledoutage',
            name='tool',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AlterField(
            model_name='reservation',
            name='tool',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AlterField(
            model_name='area',
            name='welcome_message',
            field=models.TextField(blank=True, help_text='The welcome message will be displayed on the tablet login page. You can use HTML and JavaScript.', null=True),
        ),
        migrations.AlterField(
            model_name='user',
            name='physical_access_levels',
            field=models.ManyToManyField(blank=True, to='NEMO.PhysicalAccessLevel'),
        ),
        migrations.AlterModelOptions(
            name='areaaccessrecord',
            options={},
        ),
        migrations.AlterField(
            model_name='scheduledoutage',
            name='resource',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Resource'),
        ),
        migrations.AddIndex(
            model_name='area',
            index=models.Index(fields=['name'], name='NEMO_area_name_1e7670_idx'),
        ),
        migrations.AddIndex(
            model_name='areaaccessrecord',
            index=models.Index(fields=['end'], name='NEMO_areaac_end_fae061_idx'),
        ),
        migrations.AlterField(
            model_name='areaaccessrecord',
            name='area',
            field=mptt.fields.TreeForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area'),
        ),
        migrations.AlterField(
            model_name='door',
            name='area',
            field=mptt.fields.TreeForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='doors', to='NEMO.Area'),
        ),
        migrations.AlterField(
            model_name='physicalaccesslevel',
            name='area',
            field=mptt.fields.TreeForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='_requires_area_access',
            field=mptt.fields.TreeForeignKey(blank=True, db_column='requires_area_access_id', help_text='Indicates that this tool is physically located in a billable area and requires an active area access record in order to be operated.', null=True, on_delete=django.db.models.deletion.PROTECT, to='NEMO.Area'),
        ),
        migrations.RunPython(rebuild_area_tree),
    ]



from datetime import timedelta

from django.conf import settings
from django.utils import timezone


def migration_format_datetime(universal_time):
    local_time = universal_time.astimezone(timezone.get_current_timezone())
    day = int(local_time.strftime("%d"))
    if 4 <= day <= 20 or 24 <= day <= 30:
        suffix = "th"
    else:
        suffix = ["st", "nd", "rd"][day % 10 - 1]
    return (
        local_time.strftime("%A, %B ")
        + str(day)
        + suffix
        + local_time.strftime(", %Y @ ")
        + local_time.strftime("%I:%M %p").lstrip("0")
    )


def create_news_for_version(apps, version, extra_content=None):
    if getattr(settings, "NEW_VERSION_NEWS", True):
        News = apps.get_model("NEMO", "News")
        Notification = apps.get_model("NEMO", "Notification")
        User = apps.get_model("NEMO", "User")
        news_content_type = apps.get_model("contenttypes", "ContentType").objects.get_for_model(News)
        now = timezone.now()
        story = News()
        story.title = f"What's new in NEMO {version}?"
        content = f"Thank you for updating to NEMO {version}.\n"
        if extra_content:
            content += extra_content + "\n"
        content += (
            f"\nClick on the following links to consult the <a href='https://github.com/usnistgov/NEMO/releases/tag/{version}' target='_blank'>Release Notes</a> "
            f"and the <a href='https://github.com/usnistgov/NEMO/raw/{version}/documentation/NEMO_Feature_Manual.pdf' target='_blank'>Feature manual</a>"
        )
        content = f"Originally published on {migration_format_datetime(now)} by NEMO:\n" + content.strip()
        story.original_content = content
        story.created = now
        story.all_content = content
        story.last_updated = now
        story.last_update_content = content
        story.update_count = 0
        story.save()
        users = User.objects.filter(is_active=True)
        expiration = now + timedelta(days=30)  # Unread news story notifications always expire after 30 days
        for u in users:
            Notification.objects.create(
                user=u, expiration=expiration, content_type=news_content_type, object_id=story.id
            )



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.views.decorators.http import require_GET

from NEMO.models import ContactInformation, ContactInformationCategory


@login_required
@require_GET
def contact_staff(request):
	dictionary = {
		'categories': ContactInformationCategory.objects.all(),
		'people': ContactInformation.objects.all(),
	}
	return render(request, 'contact/contact_staff.html', dictionary)
